from math import sqrt
import unittest
import io

from vamb.benchmark import Contig, Genome, Reference, Binning

REFERENCE_DICT = {
    "genomes": {
        "gA": {
            "subjA1": [
                100,
                {
                    "sA1c1": [1, 50],
                    "sA1c2": [15, 65],
                    "sA1c3": [20, 40],
                    "sA1c4": [70, 100],
                },
            ]
        },
        "gB": {
            "subjB1": [
                20,
                {
                    "sB1c1": [1, 20],
                },
            ],
            "subjB2": [
                60,
                {
                    "sB2c1": [1, 60],
                    "sB2c2": [10, 50],
                    "sB2c3": [15, 40],
                    "sB2c4": [20, 30],
                },
            ],
        },
        "gC": {
            "subjY1": [
                40,
                {
                    "sY1c1": [10, 20],
                    "sY1c2": [21, 40],
                },
            ],
            "subjY2": [
                100,
                {
                    "sY2c1": [1, 60],
                    "sY2c2": [5, 70],
                    "sY2c3": [50, 80],
                    "sY2c4": [60, 99],
                    "sY2c5": [99, 100],
                },
            ],
            "subjY3": [
                100,
                {
                    "sY3c1": [40, 50],
                    "sY3c2": [45, 55],
                },
            ],
        },
    },
    "taxmaps": [{"gA": "D", "gB": "D", "gC": "E"}, {"D": "F", "E": "F"}],
}

# C1: All of strainA
# C2: All of B2, but missing redundant contig c3
# C3: Y1
# C4: Y2 + Y3 + B1
# C5: The missing contig from C2
BINNING_STR = """C1\tsA1c1
C1\tsA1c2
C1\tsA1c3
C1\tsA1c4
C2\tsB2c1
C2\tsB2c2
C2\tsB2c4
C3\tsY1c1
C3\tsY1c2
C4\tsB1c1
C4\tsY2c1
C4\tsY2c3
C4\tsY2c4
C4\tsY2c5
C4\tsY3c1
C4\tsY3c2
C5\tsB2c3"""

# Test breadth of genomes
# Test breadth of bins
#


class TestContig(unittest.TestCase):
    def test_neg_len(self):
        with self.assertRaises(ValueError):
            Contig("x", "y", 5, 5)

        with self.assertRaises(ValueError):
            Contig("x", "y", 5, -10)

        with self.assertRaises(ValueError):
            Contig("x", "y", -1, 5)

    def test_works(self):
        c = Contig("x", "y", 9, 19)
        self.assertEqual(repr(c), 'Contig("x", "y", 9, 19)')
        c2 = Contig.subjectless("x", 10)
        self.assertEqual(c.name, c2.name)
        self.assertEqual(len(c), len(c2))

    def test_hash_eq(self):
        c1 = Contig("x", "y", 1, 2)
        c2 = Contig("y", "y", 1, 2)
        c3 = Contig("x", "other", 999, 5000)
        self.assertEqual(c1, c3)
        self.assertNotEqual(c1, c2)
        self.assertEqual({c1}, {c3})
        self.assertEqual(len({c1, c3}), 1)


class TestGenome(unittest.TestCase):
    def test_genome(self):
        g = Genome("gen")
        with self.assertRaises(ValueError):
            g.add("foo", 0)

        with self.assertRaises(ValueError):
            g.add("foo", -1)

        self.assertEqual(g.breadth, 0)
        g.add("bar", 10)
        self.assertEqual(g.breadth, 10)
        g.add("qux", 21)
        self.assertEqual(g.breadth, 31)
        self.assertEqual(repr(g), 'Genome("gen")')

        with self.assertRaises(ValueError):
            g.add("bar", 5)

        self.assertEqual(g.breadth, sum(g.sources.values()))

    def test_eq(self):
        g1 = Genome("x")
        g1.add("foo", 5)
        g1.add("bar", 10)
        g2 = Genome("x")
        g3 = Genome("different")
        self.assertEqual(g1, g2)
        self.assertNotEqual(g2, g3)
        self.assertEqual({g1}, {g2})
        self.assertEqual(len({g1, g2}), 1)


class TestBenchmark(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.reference = Reference.from_dict(REFERENCE_DICT)
        cls.binning = Binning.from_file(io.StringIO(BINNING_STR), cls.reference)

    def test_bin_basics(self):
        bins = {b.name: b for b in self.binning.bins}
        self.assertEqual(repr(bins["C1"]), 'Bin("C1")')

        self.assertEqual(bins["C1"].ncontigs, 4)
        self.assertEqual(bins["C2"].ncontigs, 3)
        self.assertEqual(bins["C3"].ncontigs, 2)
        self.assertEqual(bins["C4"].ncontigs, 7)
        self.assertEqual(bins["C5"].ncontigs, 1)

        self.assertEqual(bins["C1"].breadth, 96)
        self.assertEqual(bins["C2"].breadth, 60)
        self.assertEqual(bins["C3"].breadth, 31)
        self.assertEqual(bins["C4"].breadth, 136)
        self.assertEqual(bins["C5"].breadth, 26)

    def test_bin_confusion_matrix(self):
        bins = {b.name: b for b in self.binning.bins}
        genomes = {g.name: g for g in self.reference.genomes}
        self.assertEqual(bins["C1"].confusion_matrix(genomes["gA"]), (96, 0, 4))
        self.assertEqual(bins["C4"].confusion_matrix(genomes["gC"]), (116, 20, 124))
        self.assertEqual(
            bins["C2"].confusion_matrix(genomes["gA"]),
            (0, bins["C2"].breadth, genomes["gA"].breadth),
        )

    def test_bin_recall_precision(self):
        bins = {b.name: b for b in self.binning.bins}
        genomes = {g.name: g for g in self.reference.genomes}
        self.assertEqual(bins["C1"].recall_precision(genomes["gA"]), (0.96, 1.0))
        self.assertEqual(
            bins["C4"].recall_precision(genomes["gC"]), (116 / 240, 116 / 136)
        )
        self.assertEqual(bins["C2"].recall_precision(genomes["gA"]), (0.0, 0.0))

    def test_bin_fscore(self):
        bins = {b.name: b for b in self.binning.bins}
        genomes = {g.name: g for g in self.reference.genomes}
        pairs = [("C1", "gA"), ("C2", "gA"), ("C4", "gC")]

        for binname, gname in pairs:
            self.assertEqual(
                bins[binname].f1(genomes[gname]),
                bins[binname].fscore(1.0, genomes[gname]),
            )

        # Alternatively, these could be considered to be undefined
        self.assertEqual(bins["C1"].fscore(2.0, genomes["gB"]), 0.0)
        self.assertEqual(bins["C5"].fscore(0.5, genomes["gA"]), 0.0)

        self.assertAlmostEqual(
            bins["C4"].fscore(sqrt(2), genomes["gC"]), 0.5649350649350648
        )
        self.assertAlmostEqual(
            bins["C1"].fscore(0.4, genomes["gA"]), 0.9942857142857143
        )

    def test_ref_errors(self):
        # Can't add same contig twice
        genome = self.reference.genomeof[self.reference.contig_by_name["sA1c4"]]
        with self.assertRaises(ValueError):
            self.reference._add_contig(Contig("sA1c4", "subjA1", 10, 20), genome)

        # Can't add genome of same name
        with self.assertRaises(ValueError):
            self.reference._add_genome(Genome("gB"))

        # Can't add contig with unknown source
        with self.assertRaises(ValueError):
            self.reference._add_contig(
                Contig("newcontig", "newsubject", 10, 20), genome
            )

        # Can't add contig with unknown genome
        with self.assertRaises(ValueError):
            g = Genome("newgenome")
            g.add("subjA1", 99)
            self.reference._add_contig(Contig("newcontig", "subjA1", 10, 20), g)

        # Can't add contig longer than its subject
        with self.assertRaises(IndexError):
            ref = Reference()
            g = Genome("newgenome")
            g.add("subj", 5)
            ref._add_genome(g)
            ref._add_contig(Contig("x", "subj", 1, 6), g)

        with self.assertRaises(ValueError):
            g = Genome("newgenome")
            g.add("foo", 99)
            self.reference._add_contig(
                Contig("newcontig", "foo", 10, 20), Genome("newgenome")
            )

        # Can't add taxonomy to wrong clade
        with self.assertRaises(ValueError):
            self.reference._add_taxonomy(1, "D", "H")

        # Can't add taxonomy of a genome that does not exist
        with self.assertRaises(KeyError):
            self.reference._add_taxonomy(1, "X", "H")

    def test_ref_basics(self):
        self.assertEqual(self.reference.ngenomes, 3)
        self.assertEqual(self.reference.ncontigs, 18)
        self.assertEqual(self.reference.nranks, 3)
        self.assertEqual(
            repr(self.reference), "<Reference with 3 genomes, 18 contigs and 3 ranks>"
        )

    def test_ref_roundtrip(self):
        ref = self.reference
        buffer = io.StringIO()
        ref.save(buffer)
        buffer.seek(0)
        ref2 = Reference.from_file(buffer)

        self.assertEqual(ref.ncontigs, ref2.ncontigs)
        self.assertEqual(ref.ngenomes, ref2.ngenomes)
        self.assertEqual(ref.nranks, ref2.nranks)
        self.assertEqual(ref.genomes, ref2.genomes)
        self.assertEqual(ref.contig_by_name, ref2.contig_by_name)
        self.assertEqual(ref.taxmaps, ref2.taxmaps)

    def test_binning_basics(self):
        self.assertEqual(self.binning.nbins, 5)
        self.assertRegex(
            repr(self.binning), "<Binning with 5 bins and reference 0x[a-f0-9]+>"
        )

    def test_binning_badargs(self):
        ref = self.reference
        with self.assertRaises(ValueError):
            Binning.from_file(io.StringIO(BINNING_STR), ref, recalls=[])

        with self.assertRaises(ValueError):
            Binning.from_file(io.StringIO(BINNING_STR), ref, recalls=[-0.01, 0.5])

        with self.assertRaises(ValueError):
            Binning.from_file(io.StringIO(BINNING_STR), ref, recalls=[0.5, 0.5])

    def test_binning_disjoint(self):
        s = BINNING_STR + "\nC99\tsY2c1"
        buffer = io.StringIO(s)

        with self.assertRaises(ValueError):
            Binning.from_file(buffer, self.reference)

        buffer.seek(0)
        # Does not raise an error
        self.assertIsInstance(
            Binning.from_file(buffer, self.reference, disjoint=False), Binning
        )

    def test_binning_strain_counter(self):
        # This approach is simple and easy to verify to be correct, but very inefficient.
        # Strain-level counter (level = 0)
        for (min_recall, min_precision), n_obs in self.binning.counters[0].items():
            n_exp = 0
            for genome in self.reference.genomes:
                for bin in self.binning.bins:
                    recall, precision = bin.recall_precision(genome)
                    if recall >= min_recall and precision >= min_precision:
                        n_exp += 1
                        break

            self.assertEqual(n_exp, n_obs)

    def test_binning_clade_counter(self):
        counters: list[
            dict[tuple[float, float], int]
        ] = self.binning.counters  # type: ignore
        genomesof = [{"D": ["gA", "gB"], "E": ["gC"]}, {"F": ["gA", "gB", "gC"]}]

        genomes = {g.name: g for g in self.reference.genomes}
        for rank in (1, 2):
            for (min_recall, min_precision), n_obs in counters[rank].items():
                seen = {c: False for c in genomesof[rank - 1]}
                for clade, genomenames in genomesof[rank - 1].items():
                    for bin in self.binning.bins:
                        rec = 0.0
                        prec = 0.0
                        for genomename in genomenames:
                            recall, precision = bin.recall_precision(
                                genomes[genomename]
                            )
                            rec = max(rec, recall)
                            prec += precision
                        if rec >= min_recall and prec >= min_precision:
                            seen[clade] = True

                self.assertEqual(sum(seen.values()), n_obs)

    def test_filtering(self):
        def test_binnames(bin: Binning, names):
            self.assertEqual({b.name for b in bin.bins}, set(names))

        bin2 = Binning.from_file(
            io.StringIO(BINNING_STR), self.reference, mincontigs=0, minsize=0
        )
        test_binnames(bin2, ["C1", "C2", "C3", "C4", "C5"])

        bin2 = Binning.from_file(io.StringIO(BINNING_STR), self.reference, mincontigs=3)
        test_binnames(bin2, ["C1", "C2", "C4"])

        bin2 = Binning.from_file(io.StringIO(BINNING_STR), self.reference, minsize=50)
        test_binnames(bin2, ["C1", "C2", "C4"])

        bin2 = Binning.from_file(io.StringIO(BINNING_STR), self.reference, minsize=100)
        test_binnames(bin2, ["C4"])

        bin2 = Binning.from_file(
            io.StringIO(BINNING_STR), self.reference, minsize=50, mincontigs=4
        )
        test_binnames(bin2, ["C1", "C4"])

    def test_printmatrix(self):
        with self.assertRaises(IndexError):
            self.binning.print_matrix(3)

        buffer = io.StringIO()
        self.binning.print_matrix(0, buffer)



import io
import os
import unittest
import tempfile
import gzip
import bz2
import lzma
import random
import itertools
import numpy as np
import string
import torch

import vamb
import testtools

PARENTDIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DATADIR = os.path.join(PARENTDIR, "test", "data")


class TestReader(unittest.TestCase):
    def setUp(self):
        # Do not create files if they exist
        if hasattr(self, "files"):
            return None

        self.files = []
        with open(os.path.join(DATADIR, "fasta.fna"), "rb") as file:
            self.lines = list(file)

        for f in [gzip.compress, bz2.compress, lzma.compress, lambda x: x]:
            file = tempfile.NamedTemporaryFile()
            file.write(f(b"".join(self.lines)))
            file.flush()
            self.files.append(file)

        return None

    def test_reader(self):
        for file in self.files:
            with vamb.vambtools.Reader(file.name) as reader:
                self.assertEqual(self.lines, list(reader))


class TestPushArray(unittest.TestCase):
    def test_append(self):
        arr = vamb.vambtools.PushArray(np.float32, start_capacity=3)
        values = [1.0, 1, -3, 55.3]
        for i in values:
            arr.append(i)

        self.assertEqual(len(arr), 4)
        self.assertTrue(np.all(arr.take() == np.array(values, dtype=np.float32)))
        arr.clear()
        self.assertEqual(len(arr), 0)

    def test_extend(self):
        arr = vamb.vambtools.PushArray(np.float32, start_capacity=1)
        values = [1.0, 1, -3, 55.3, 0, 53, -1.2]
        arr.extend(values)
        self.assertTrue(np.all(arr.take() == np.array(values, dtype=np.float32)))

        arr = vamb.vambtools.PushArray(np.float32, start_capacity=1)
        for i in range(1, 10000, 100):
            arr.extend(range(i, i + 100))
        self.assertTrue(
            np.all(arr.take() == np.array(range(1, 10001), dtype=np.float32))
        )

    def test_clear(self):
        arr = vamb.vambtools.PushArray(np.float32, start_capacity=1)
        values = [1.0, 1, -3, 55.3, 0, 53, -1.2]
        arr.extend(values)
        self.assertTrue(np.all(arr.take() == np.array(values, dtype=np.float32)))
        arr.clear()
        self.assertTrue(len(arr.take()) == 0)


class TestFASTAEntry(unittest.TestCase):
    def test_init(self):
        passing = [b"0", b"foo", b"FOO123_@@^", b"%%;;/=~~"]
        failing = [
            b"",  # empty
            b">",  # empty
            b">foo",  # with >
            b"=",  # cannot start with =
            b"*",  # cannot start with *
            b"\xff",  # out of range
            b"))",
            b" ",  # zero-length identifier
        ]

        for i in passing:
            x = vamb.vambtools.FastaEntry(i, bytearray(b"TAG"))
            self.assertTrue(i.decode(), x.header)

        for i in failing:
            with self.assertRaises(ValueError):
                vamb.vambtools.FastaEntry(i, bytearray(b"TAG"))

        # Various
        with self.assertRaises(ValueError):
            vamb.vambtools.FastaEntry(b"foo", bytearray(b"ATGCpN"))

    def test_masking(self):
        self.assertEqual(
            vamb.vambtools.FastaEntry(b"foo", bytearray(b"TaGkmYnAC")).sequence,
            bytearray(b"TaGkmYnAC"),
        )
        self.assertEqual(
            vamb.vambtools.FastaEntry(b"foo", bytearray()).sequence, bytearray()
        )

    def test_various(self):
        # Length
        self.assertEqual(len(vamb.vambtools.FastaEntry(b"x", bytearray(b"TAGCA"))), 5)
        self.assertEqual(len(vamb.vambtools.FastaEntry(b"x", bytearray())), 0)
        self.assertEqual(
            len(vamb.vambtools.FastaEntry(b"x", bytearray(b"TGTAmnyAncC"))), 11
        )

        # format
        self.assertEqual(
            vamb.vambtools.FastaEntry(b"x", bytearray(b"TAGCA")).format(), ">x\nTAGCA"
        )
        self.assertEqual(
            vamb.vambtools.FastaEntry(b"yz", bytearray()).format(), ">yz\n"
        )
        self.assertEqual(
            vamb.vambtools.FastaEntry(b"1_2", bytearray(b"TGTAmnyAncC")).format(),
            ">1_2\nTGTAmnyAncC",
        )

    def test_kmercounts(self):
        seq = vamb.vambtools.FastaEntry(b"X", bytearray(b"TTAyCAAnGAC"))

        with self.assertRaises(ValueError):
            seq.kmercounts(0)

        with self.assertRaises(ValueError):
            seq.kmercounts(13)

        self.assertTrue(np.all(seq.kmercounts(1) == np.array([4, 2, 1, 2])))

        self.assertTrue(
            np.all(
                seq.kmercounts(2)
                == np.array(
                    [
                        1,
                        1,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        0,
                        1,
                    ]
                )
            )
        )

    def test_random_kmercounts(self):
        indexof = {
            "".join(ncs): idx
            for (idx, ncs) in enumerate(itertools.product("ACGT", repeat=4))
        }
        seq = testtools.make_randseq(random.Random(), 900, 1100)
        sequence = seq.sequence.decode()
        manual_counts = np.zeros(256, dtype=int)
        for i in range(len(sequence) - 3):
            ind = indexof.get(sequence[i : i + 4].upper())
            if ind is not None:
                manual_counts[ind] += 1

        automatic = seq.kmercounts(4)
        self.assertTrue(np.all(manual_counts == automatic))

    def test_rename(self):
        seq = vamb.vambtools.FastaEntry(b"foo", bytearray(b"TaGkmYnAC"))

        # Does not error
        seq.rename(b"identifier\t desc")
        self.assertEqual(seq.identifier, "identifier")
        self.assertEqual(seq.description, "\t desc")

        seq.rename(b"newname")
        self.assertEqual(seq.identifier, "newname")
        self.assertEqual(seq.description, "")

        # Errors
        with self.assertRaises(ValueError):
            seq.rename(b"\tabc def")

        with self.assertRaises(ValueError):
            seq.rename(b"=123")

        with self.assertRaises(ValueError):
            seq.rename(b"")

        with self.assertRaises(ValueError):
            seq.rename(b"\xff")


class TestFASTAReader(unittest.TestCase):
    def test_bad_files(self):
        # First non-comment line must be header
        with self.assertRaises(ValueError):
            data = b"#foo\n#bar\n  \n>foo\nTAG".splitlines()
            list(vamb.vambtools.byte_iterfasta(data))

        # String input
        with self.assertRaises(TypeError):
            data = ">abc\nTAG\na\nAC".splitlines()
            list(vamb.vambtools.byte_iterfasta(data))

    # Various correct formats
    def test_good_files(self):
        # Empty file
        data = b"".splitlines()
        records = list(vamb.vambtools.byte_iterfasta(data))
        self.assertEqual(0, len(records))

        # Only comments
        data = b"#bar\n#foo\n#".splitlines()
        records = list(vamb.vambtools.byte_iterfasta(data))
        self.assertEqual(0, len(records))

        # A few sequences
        data = b"#bar\n#foo\n>ab\nTA\nT\n \t\nA\nNN\n>foo\nCyAmmkg\n>bar\n".splitlines()
        records = list(vamb.vambtools.byte_iterfasta(data))
        self.assertEqual(3, len(records))
        self.assertEqual(records[0].sequence, bytearray(b"TATANN"))
        self.assertEqual(records[1].sequence, bytearray(b"CyAmmkg"))
        self.assertEqual(records[2].sequence, bytearray(b""))

        # Comment inside seqs
        data = b"#bar\n>foo\nTAG\n##xxx\n>@@AA\nAAA".splitlines()
        records = list(vamb.vambtools.byte_iterfasta(data))
        self.assertEqual(records[0].sequence, bytearray(b"TAG"))


class TestZscore(unittest.TestCase):
    arr = np.array([[1, 2, 2.5], [2, 4, 3], [0.9, 3.1, 2.8]])
    zscored = np.array(
        [
            [-1.44059316, -0.3865006, 0.14054567],
            [-0.3865006, 1.7216845, 0.66759195],
            [-1.54600241, 0.77300121, 0.45677344],
        ]
    )

    def almost_similar(self, x, y):
        self.assertTrue(np.all(np.abs(x - y) < 1e-6))

    def test_simple(self):
        self.almost_similar(vamb.vambtools.zscore(self.arr), self.zscored)

    def test_axes(self):
        self.almost_similar(
            vamb.vambtools.zscore(self.arr, axis=0),
            np.array(
                [
                    [-0.60404045, -1.26346568, -1.29777137],
                    [1.40942772, 1.18195176, 1.13554995],
                    [-0.80538727, 0.08151391, 0.16222142],
                ]
            ),
        )

        self.almost_similar(
            vamb.vambtools.zscore(self.arr, axis=1),
            np.array(
                [
                    [-1.33630621, 0.26726124, 1.06904497],
                    [-1.22474487, 1.22474487, 0.0],
                    [-1.40299112, 0.85548239, 0.54750873],
                ]
            ),
        )

    def test_axis_bounds(self):
        with self.assertRaises(np.AxisError):
            vamb.vambtools.zscore(self.arr, axis=-1)

        with self.assertRaises(np.AxisError):
            vamb.vambtools.zscore(self.arr, axis=2)

    def test_integer(self):
        with self.assertRaises(TypeError):
            vamb.vambtools.zscore(np.array([1, 2, 3]), inplace=True)

    def test_novar(self):
        arr = np.array([4, 4, 4])
        self.almost_similar(vamb.vambtools.zscore(arr), np.array([0, 0, 0]))

    def test_inplace(self):
        vamb.vambtools.zscore(self.arr, inplace=True)
        self.almost_similar(self.arr, self.zscored)


class TestInplaceMaskArray(unittest.TestCase):
    def almost_similar_np(self, x, y):
        self.assertTrue(np.all(np.abs(x - y) < 1e-6))

    def almost_similar_torch(self, x, y):
        self.assertTrue(torch.all(torch.abs(x - y) < 1e-6))

    def test_numpy(self):
        arr = np.random.random((10, 3)).astype(np.float32)
        mask = np.array([0, 1, 1, 1, 1, 0, 0, 0, 0, 0]).astype(np.uint8)
        arr2 = arr[[bool(i) for i in mask]]
        vamb.vambtools.numpy_inplace_maskarray(arr, mask)
        self.almost_similar_np(arr, arr2)

        # arr is now too short after being masked
        with self.assertRaises(ValueError):
            vamb.vambtools.numpy_inplace_maskarray(arr, mask)

    def test_torch(self):
        arr = torch.rand(10, 3)
        mask = torch.tensor([0, 1, 1, 1, 1, 0, 0, 0, 0, 0], dtype=bool)
        arr2 = arr[[bool(i) for i in mask]]
        vamb.vambtools.torch_inplace_maskarray(arr, mask)
        self.almost_similar_torch(arr, arr2)

        # arr is now too short after being masked
        with self.assertRaises(ValueError):
            vamb.vambtools.torch_inplace_maskarray(arr, mask)


class TestHashRefNames(unittest.TestCase):
    def test_refhash(self):
        names = ["foo", "9", "eleven", "a"]
        b1 = vamb.vambtools.hash_refnames(names)
        names[1] = names[1] + "x"
        b2 = vamb.vambtools.hash_refnames(names)
        names[1] = names[1][:-1] + " \t"  # it strips whitespace off right end
        b3 = vamb.vambtools.hash_refnames(names)
        names = names[::-1]
        b4 = vamb.vambtools.hash_refnames(names)
        names = (i + "   " for i in names[::-1])
        b5 = vamb.vambtools.hash_refnames(names)
        b6 = vamb.vambtools.hash_refnames(names)  # now empty generator
        b7 = vamb.vambtools.hash_refnames([])

        self.assertNotEqual(b1, b2)
        self.assertEqual(b1, b3)
        self.assertNotEqual(b1, b4)
        self.assertEqual(b1, b5)
        self.assertNotEqual(b1, b6)
        self.assertEqual(b6, b7)


class TestBinSplit(unittest.TestCase):
    before = [("bin1", ["s1-c1", "s1-c2", "s8-c1", "s9-c11"]), ("bin2", ["s12-c0"])]

    after = [
        ("s1-bin1", {"s1-c1", "s1-c2"}),
        ("s8-bin1", {"s8-c1"}),
        ("s9-bin1", {"s9-c11"}),
        ("s12-bin2", {"s12-c0"}),
    ]

    def test_split(self):
        self.assertEqual(list(vamb.vambtools.binsplit(self.before, "-")), self.after)

    def test_badsep(self):
        with self.assertRaises(KeyError):
            list(vamb.vambtools.binsplit(self.before, "2"))

    def test_badtype(self):
        with self.assertRaises(TypeError):
            list(vamb.vambtools.binsplit([(1, [2])], ""))


class TestWriteClusters(unittest.TestCase):
    test_clusters = [("C1", set("abc")), ("C2", set("hkjlmn")), ("C3", set("x"))]
    io = io.StringIO()

    def setUp(self):
        self.io.truncate(0)
        self.io.seek(0)

    def linesof(self, str):
        return list(filter(lambda x: not x.startswith("#"), str.splitlines()))

    def conforms(self, str, clusters):
        lines = self.linesof(str)
        self.assertEqual(len(lines), sum(len(v) for (k, v) in clusters))
        allcontigs = set()
        printed = set()
        printed_names = set()
        read = set()
        read_names = set()

        for k, v in clusters:
            allcontigs.update(v)

        for line in lines:
            name, _, contig = line.partition("\t")
            printed.add(contig)
            printed_names.add(name)

        for k, v in vamb.vambtools.read_clusters(io.StringIO(str)).items():
            read.update(v)
            read_names.add(k)

        self.assertEqual(allcontigs, printed)
        self.assertEqual(allcontigs, read)
        self.assertEqual(read_names, printed_names)

    def test_not_writable(self):
        buf = io.BufferedReader(io.BytesIO(b""))
        with self.assertRaises(ValueError):
            vamb.vambtools.write_clusters(buf, self.test_clusters)

    def test_invalid_max_clusters(self):
        with self.assertRaises(ValueError):
            vamb.vambtools.write_clusters(self.io, self.test_clusters, max_clusters=0)

        with self.assertRaises(ValueError):
            vamb.vambtools.write_clusters(self.io, self.test_clusters, max_clusters=-11)

    def test_header_has_newline(self):
        with self.assertRaises(ValueError):
            vamb.vambtools.write_clusters(self.io, self.test_clusters, header="foo\n")

    def test_normal(self):
        vamb.vambtools.write_clusters(self.io, self.test_clusters, header="someheader")
        self.assertTrue(self.io.getvalue().startswith("# someheader"))
        self.conforms(self.io.getvalue(), self.test_clusters)

    def test_max_clusters(self):
        vamb.vambtools.write_clusters(self.io, self.test_clusters, max_clusters=2)
        lines = self.linesof(self.io.getvalue())
        self.assertEqual(len(lines), 9)
        self.conforms(self.io.getvalue(), self.test_clusters[:2])

    def test_min_size(self):
        vamb.vambtools.write_clusters(self.io, self.test_clusters, min_size=5)
        lines = self.linesof(self.io.getvalue())
        self.assertEqual(len(lines), 6)
        self.conforms(self.io.getvalue(), self.test_clusters[1:2])


class TestWriteBins(unittest.TestCase):
    file = io.BytesIO()
    N_BINS = 10
    minsize = 5 * 175  # mean of bin size

    @classmethod
    def setUpClass(cls):
        bins: dict[str, set[str]] = dict()
        seqs: dict[str, vamb.vambtools.FastaEntry] = dict()
        for i in range(cls.N_BINS):
            binname = "".join(random.choices(string.ascii_letters, k=12))
            bins[binname] = set()
            for j in range(random.randrange(3, 7)):
                seq = testtools.make_randseq(random.Random(), 100, 250)
                seqs[seq.identifier] = seq
                bins[binname].add(seq.identifier)
                cls.file.write(seq.format().encode())
                cls.file.write(b"\n")

        cls.bins = bins
        cls.seqs = seqs

    def setUp(self):
        self.file.seek(0)

    def test_bad_params(self):
        # Too many bins for maxbins
        with self.assertRaises(ValueError):
            vamb.vambtools.write_bins(
                "foo", self.bins, self.file, maxbins=self.N_BINS - 1
            )

        # Negative minsize
        with self.assertRaises(ValueError):
            vamb.vambtools.write_bins(
                "foo", self.bins, self.file, maxbins=self.N_BINS + 1, minsize=-1
            )

        # Parent does not exist
        with self.assertRaises(NotADirectoryError):
            vamb.vambtools.write_bins(
                "svogew/foo", self.bins, self.file, maxbins=self.N_BINS + 1
            )

        # Target file already exists
        with self.assertRaises(FileExistsError):
            with tempfile.NamedTemporaryFile() as file:
                vamb.vambtools.write_bins(
                    file.name, self.bins, self.file, maxbins=self.N_BINS + 1
                )

        # One contig missing from fasta dict
        with self.assertRaises(IndexError):
            bins = {k: v.copy() for k, v in self.bins.items()}
            next(iter(bins.values())).add("a_new_bin_which_does_not_exist")
            vamb.vambtools.write_bins("foo", bins, self.file, maxbins=self.N_BINS + 1)

    def test_round_trip(self):
        with tempfile.TemporaryDirectory() as dir:
            vamb.vambtools.write_bins(
                dir, self.bins, self.file, maxbins=self.N_BINS, minsize=self.minsize
            )

            reconstructed_bins: dict[str, set[str]] = dict()
            for filename in os.listdir(dir):
                with open(os.path.join(dir, filename), "rb") as file:
                    entries = list(vamb.vambtools.byte_iterfasta(file))
                    if sum(map(len, entries)) < self.minsize:
                        continue
                    binname = filename[:-4]
                    reconstructed_bins[binname] = set()
                    for entry in entries:
                        reconstructed_bins[binname].add(entry.identifier)

        filtered_bins = {
            k: v
            for (k, v) in self.bins.items()
            if sum(len(self.seqs[vi]) for vi in v) >= self.minsize
        }

        # Same bins
        self.assertEqual(len(filtered_bins), len(reconstructed_bins))
        self.assertEqual(
            sum(map(len, filtered_bins.values())),
            sum(map(len, reconstructed_bins.values())),
        )
        self.assertEqual(filtered_bins, reconstructed_bins)



#!/usr/bin/env python3

# More imports below, but the user's choice of processors must be parsed before
# numpy can be imported.
import vamb
import numpy as np
import sys
import os
import argparse
import torch
import datetime
import time
from math import isfinite
from typing import Optional, IO
from pathlib import Path
from collections.abc import Sequence
from torch.utils.data import DataLoader

_ncpu = os.cpu_count()
DEFAULT_THREADS = 8 if _ncpu is None else min(_ncpu, 8)

# These MUST be set before importing numpy
# I know this is a shitty hack, see https://github.com/numpy/numpy/issues/11826
os.environ["MKL_NUM_THREADS"] = str(DEFAULT_THREADS)
os.environ["NUMEXPR_NUM_THREADS"] = str(DEFAULT_THREADS)
os.environ["OMP_NUM_THREADS"] = str(DEFAULT_THREADS)

# Append vamb to sys.path to allow vamb import even if vamb was not installed
# using pip
parentdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(parentdir)


def try_make_dir(name):
    try:
        os.mkdir(name)
    except FileExistsError:
        pass


class FASTAPath(type(Path())):
    pass


class CompositionPath(type(Path())):
    pass


class CompositionOptions:
    __slots__ = ["path", "min_contig_length"]

    def __init__(
        self, fastapath: Optional[Path], npzpath: Optional[Path], min_contig_length: int
    ):
        assert isinstance(fastapath, (Path, type(None)))
        assert isinstance(npzpath, (Path, type(None)))
        assert isinstance(min_contig_length, int)

        if min_contig_length < 250:
            raise argparse.ArgumentTypeError(
                "Minimum contig length must be at least 250"
            )

        if not (fastapath is None) ^ (npzpath is None):
            raise argparse.ArgumentTypeError(
                "Must specify either FASTA or composition path"
            )

        for path in (fastapath, npzpath):
            if path is not None and not path.is_file():
                raise FileNotFoundError(path)

        if fastapath is not None:
            self.path = FASTAPath(fastapath)
        else:
            assert npzpath is not None
            self.path = CompositionPath(npzpath)
        self.min_contig_length = min_contig_length


class AbundancePath(type(Path())):
    pass


class AbundanceOptions:
    __slots__ = ["path", "min_alignment_id", "refcheck"]

    def __init__(
        self,
        bampaths: Optional[list[Path]],
        abundancepath: Optional[Path],
        min_alignment_id: Optional[float],
        refcheck: bool,
    ):
        assert isinstance(bampaths, (list, type(None)))
        assert isinstance(abundancepath, (Path, type(None)))
        assert isinstance(min_alignment_id, (float, type(None)))
        assert isinstance(refcheck, bool)

        # Make sure only one RPKM input is there
        if not (bampaths is not None) + (abundancepath is not None) == 1:
            raise argparse.ArgumentTypeError(
                "Must specify exactly one of BAM files or abundance NPZ file input"
            )

        if abundancepath is not None:
            if not abundancepath.is_file():
                raise FileNotFoundError(
                    f'Not an existing non-directory file: "{str(abundancepath)}"'
                )
            self.path = AbundancePath(abundancepath)

        elif bampaths is not None:
            for bampath in bampaths:
                if not bampath.is_file():
                    raise FileNotFoundError(
                        f'Not an existing non-directory file: "{str(bampath)}"'
                    )
            self.path = bampaths

        if min_alignment_id is not None:
            if bampaths is None:
                raise argparse.ArgumentTypeError(
                    "If minid is set, RPKM must be passed as bam files"
                )
            if (
                not isfinite(min_alignment_id)
                or min_alignment_id < 0.0
                or min_alignment_id > 1.0
            ):
                raise argparse.ArgumentTypeError(
                    "Minimum nucleotide ID must be in [0,1]"
                )
            self.min_alignment_id = min_alignment_id
        else:
            self.min_alignment_id = 0.0

        self.refcheck = refcheck


class VAEOptions:
    __slots__ = ["nhiddens", "nlatent", "beta", "dropout"]

    def __init__(
        self,
        nhiddens: Optional[list[int]],
        nlatent: int,
        beta: float,
        dropout: Optional[float],
    ):
        assert isinstance(nhiddens, (list, type(None)))
        assert isinstance(nlatent, int)

        assert isinstance(beta, float)
        assert isinstance(dropout, (float, type(None)))

        if nhiddens is not None and any(i < 1 for i in nhiddens):
            raise argparse.ArgumentTypeError(
                f"Minimum 1 neuron per layer, not {min(nhiddens)}"
            )
        self.nhiddens = nhiddens

        if nlatent < 1:
            raise argparse.ArgumentTypeError(f"Minimum 1 latent neuron, not {nlatent}")
        self.nlatent = nlatent

        if beta <= 0 or not isfinite(beta):
            raise argparse.ArgumentTypeError("beta cannot be negative or zero")
        self.beta = beta

        if dropout is not None and (dropout < 0 or dropout >= 1):
            raise argparse.ArgumentTypeError("dropout must be in 0 <= d < 1")
        self.dropout = dropout


class AAEOptions:
    def __init__(
        self,
        nhiddens: int,
        nlatent_z: int,
        nlatent_y: int,
        sl: float,
        slr: float,
    ):
        assert isinstance(nhiddens, int)
        assert isinstance(nlatent_z, int)
        assert isinstance(nlatent_y, int)
        assert isinstance(sl, float)
        assert isinstance(slr, float)

        self.nhiddens = nhiddens
        self.nlatent_z = nlatent_z
        self.nlatent_y = nlatent_y
        self.sl = sl
        self.slr = slr


class EncoderOptions:
    __slots__ = ["vae_options", "aae_options", "alpha"]

    def __init__(
        self,
        vae_options: Optional[VAEOptions],
        aae_options: Optional[AAEOptions],
        alpha: Optional[float],
    ):
        assert isinstance(alpha, (float, type(None)))
        assert isinstance(vae_options, (VAEOptions, type(None)))
        assert isinstance(aae_options, (AAEOptions, type(None)))

        if alpha is not None and (alpha <= 0 or alpha) >= 1:
            raise argparse.ArgumentTypeError("alpha must be above 0 and below 1")
        self.alpha = alpha

        if vae_options is None and aae_options is None:
            raise argparse.ArgumentTypeError(
                "You must specify either VAE or AAE or both as encoders, cannot run with no encoders."
            )

        self.vae_options = vae_options
        self.aae_options = aae_options


class VAETrainingOptions:
    def __init__(self, nepochs: int, batchsize: int, batchsteps: list[int]):
        assert isinstance(nepochs, int)
        assert isinstance(batchsize, int)
        assert isinstance(batchsteps, list)

        if nepochs < 1:
            raise argparse.ArgumentTypeError(f"Minimum 1 epoch, not {nepochs}")
        self.nepochs = nepochs

        if batchsize < 1:
            raise argparse.ArgumentTypeError(f"Minimum batchsize of 1, not {batchsize}")
        self.batchsize = batchsize

        batchsteps = sorted(set(batchsteps))
        if max(batchsteps, default=0) >= self.nepochs:
            raise argparse.ArgumentTypeError("All batchsteps must be less than nepochs")

        if min(batchsteps, default=1) < 1:
            raise argparse.ArgumentTypeError("All batchsteps must be 1 or higher")
        self.batchsteps = batchsteps


class AAETrainingOptions:
    def __init__(
        self,
        nepochs: int,
        batchsize: int,
        batchsteps: list[int],
        temp: Optional[float],
    ):
        assert isinstance(nepochs, int)
        assert isinstance(batchsize, int)
        assert isinstance(batchsteps, list)

        if nepochs < 1:
            raise argparse.ArgumentTypeError(f"Minimum 1 epoch, not {nepochs}")
        self.nepochs = nepochs

        if batchsize < 1:
            raise argparse.ArgumentTypeError(f"Minimum batchsize of 1, not {batchsize}")
        self.batchsize = batchsize

        batchsteps = sorted(set(batchsteps))
        if max(batchsteps, default=0) >= self.nepochs:
            raise argparse.ArgumentTypeError("All batchsteps must be less than nepochs")

        if min(batchsteps, default=1) < 1:
            raise argparse.ArgumentTypeError("All batchsteps must be 1 or higher")
        self.batchsteps = batchsteps

        assert isinstance(temp, (float, type(None)))
        self.temp = temp


class TrainingOptions:
    def __init__(
        self,
        encoder_options: EncoderOptions,
        vae_options: Optional[VAETrainingOptions],
        aae_options: Optional[AAETrainingOptions],
        lrate: float,
    ):
        assert isinstance(lrate, float)

        assert (encoder_options.vae_options is None) == (vae_options is None)
        assert (encoder_options.aae_options is None) == (aae_options is None)

        if lrate <= 0.0:
            raise argparse.ArgumentTypeError("Learning rate must be positive")
        self.lrate = lrate

        self.vae_options = vae_options
        self.aae_options = aae_options
        self.lrate = lrate


class ClusterOptions:
    __slots__ = [
        "window_size",
        "min_successes",
        "min_cluster_size",
        "max_clusters",
        "binsplit_separator",
    ]

    def __init__(
        self,
        window_size: int,
        min_successes: int,
        min_cluster_size: int,
        max_clusters: Optional[int],
        binsplit_separator: Optional[str],
    ):
        assert isinstance(window_size, int)
        assert isinstance(min_successes, int)
        assert isinstance(min_cluster_size, int)
        assert isinstance(max_clusters, (int, type(None)))
        assert isinstance(binsplit_separator, (str, type(None)))

        if window_size < 1:
            raise argparse.ArgumentTypeError("Window size must be at least 1")
        self.window_size = window_size

        if min_cluster_size < 1:
            raise argparse.ArgumentTypeError("Minimum cluster size must be at least 0")
        self.min_cluster_size = min_cluster_size

        if min_successes < 1 or min_successes > window_size:
            raise argparse.ArgumentTypeError(
                "Minimum cluster size must be in 1:windowsize"
            )
        self.min_successes = min_successes

        if max_clusters is not None and max_clusters < 1:
            raise argparse.ArgumentTypeError("Max clusters must be at least 1")
        self.max_clusters = max_clusters

        if binsplit_separator is not None and len(binsplit_separator) == 0:
            raise argparse.ArgumentTypeError(
                "Binsplit separator cannot be an empty string"
            )
        self.binsplit_separator = binsplit_separator


class VambOptions:
    __slots__ = [
        "out_dir",
        "n_threads",
        "comp_options",
        "min_fasta_output_size",
        "seed",
        "cuda",
    ]

    def __init__(
        self,
        out_dir: Path,
        n_threads: int,
        comp_options: CompositionOptions,
        min_fasta_output_size: Optional[int],
        seed: int,
        cuda: bool,
    ):
        assert isinstance(out_dir, Path)
        assert isinstance(n_threads, int)
        assert isinstance(comp_options, CompositionOptions)
        assert isinstance(min_fasta_output_size, (int, type(None)))
        assert isinstance(seed, int)
        assert isinstance(cuda, bool)

        # Outdir does not exist
        if out_dir.exists():
            raise FileExistsError(out_dir)

        # Outdir is in an existing parent dir
        parent_dir = out_dir.parent
        if not parent_dir.is_dir():
            raise NotADirectoryError(parent_dir)
        self.out_dir = out_dir

        if n_threads < 1:
            raise ValueError(f"Must pass at least 1 thread, not {n_threads}")
        self.n_threads = n_threads

        if min_fasta_output_size is not None:
            if not isinstance(comp_options.path, FASTAPath):
                raise argparse.ArgumentTypeError(
                    "If minfasta is not None, "
                    "input fasta file must be given explicitly"
                )
            if min_fasta_output_size < 0:
                raise argparse.ArgumentTypeError(
                    "Minimum FASTA output size must be nonnegative"
                )
        self.min_fasta_output_size = min_fasta_output_size

        if cuda and not torch.cuda.is_available():
            raise ModuleNotFoundError(
                "Cuda is not available on your PyTorch installation"
            )
        self.seed = seed
        self.cuda = cuda


def log(string: str, logfile: IO[str], indent: int = 0):
    print(("\t" * indent) + string, file=logfile)
    logfile.flush()


def calc_tnf(
    options: CompositionOptions,
    outdir: Path,
    logfile: IO[str],
) -> vamb.parsecontigs.Composition:
    begintime = time.time()
    log("\nLoading TNF", logfile, 0)
    log(f"Minimum sequence length: {options.min_contig_length}", logfile, 1)

    path = options.path

    if isinstance(path, CompositionPath):
        log(f"Loading composition from npz {path}", logfile, 1)
        composition = vamb.parsecontigs.Composition.load(str(path))
        composition.filter_min_length(options.min_contig_length)
    else:
        assert isinstance(path, FASTAPath)
        log(f"Loading data from FASTA file {path}", logfile, 1)
        with vamb.vambtools.Reader(str(path)) as file:
            composition = vamb.parsecontigs.Composition.from_file(
                file, minlength=options.min_contig_length
            )
        composition.save(outdir.joinpath("composition.npz"))

    elapsed = round(time.time() - begintime, 2)
    print("", file=logfile)
    log(
        f"Kept {composition.count_bases()} bases in {composition.nseqs} sequences",
        logfile,
        1,
    )
    log(f"Processed TNF in {elapsed} seconds.", logfile, 1)

    return composition


def calc_rpkm(
    abundance_options: AbundanceOptions,
    outdir: Path,
    comp_metadata: vamb.parsecontigs.CompositionMetaData,
    nthreads: int,
    logfile: IO[str],
) -> vamb.parsebam.Abundance:
    begintime = time.time()
    log("\nLoading depths", logfile)
    log(
        f'Reference hash: {comp_metadata.refhash.hex() if abundance_options.refcheck else "None"}',
        logfile,
        1,
    )

    path = abundance_options.path
    if isinstance(path, AbundancePath):
        log(f"Loading depths from npz array {str(path)}", logfile, 1)

        abundance = vamb.parsebam.Abundance.load(
            path, comp_metadata.refhash if abundance_options.refcheck else None
        )
        # I don't want this check in any constructors of abundance, since the constructors
        # should be able to skip this check in case comp and abundance are independent.
        # But when running the main Vamb workflow, we need to assert this.
        if abundance.nseqs != comp_metadata.nseqs:
            assert not abundance_options.refcheck
            raise ValueError(
                f"Loaded abundance has {abundance.nseqs} sequences, "
                f"but composition has {comp_metadata.nseqs}."
            )
    else:
        assert isinstance(path, list)
        log(f"Parsing {len(path)} BAM files with {nthreads} threads", logfile, 1)

        abundance = vamb.parsebam.Abundance.from_files(
            path,
            outdir.joinpath("tmp").joinpath("pycoverm"),
            comp_metadata,
            abundance_options.refcheck,
            abundance_options.min_alignment_id,
            nthreads,
        )
        abundance.save(outdir.joinpath("abundance.npz"))

    log(f"Min identity: {abundance.minid}\n", logfile, 1)
    log("Order of columns is:", logfile, 1)
    log("\n\t".join(abundance.samplenames), logfile, 1)

    elapsed = round(time.time() - begintime, 2)
    print("", file=logfile)
    log(f"Processed RPKM in {elapsed} seconds.", logfile, 1)

    return abundance


def trainvae(
    vae_options: VAEOptions,
    training_options: VAETrainingOptions,
    vamb_options: VambOptions,
    lrate: float,
    alpha: Optional[float],
    data_loader: DataLoader,
    logfile: IO[str],
) -> np.ndarray:
    begintime = time.time()
    log("\nCreating and training VAE", logfile)

    nsamples = data_loader.dataset.tensors[0].shape[1]
    vae = vamb.encode.VAE(
        nsamples,
        nhiddens=vae_options.nhiddens,
        nlatent=vae_options.nlatent,
        alpha=alpha,
        beta=vae_options.beta,
        dropout=vae_options.dropout,
        cuda=vamb_options.cuda,
        seed=vamb_options.seed,
    )

    log("Created VAE", logfile, 1)
    modelpath = vamb_options.out_dir.joinpath("model.pt")
    vae.trainmodel(
        vamb.encode.set_batchsize(data_loader, training_options.batchsize),
        nepochs=training_options.nepochs,
        lrate=lrate,
        batchsteps=training_options.batchsteps,
        logfile=logfile,
        modelfile=modelpath,
    )

    print("", file=logfile)
    log("Encoding to latent representation", logfile, 1)
    latent = vae.encode(data_loader)
    vamb.vambtools.write_npz(vamb_options.out_dir.joinpath("latent.npz"), latent)
    del vae  # Needed to free "latent" array's memory references?

    elapsed = round(time.time() - begintime, 2)
    log(f"Trained VAE and encoded in {elapsed} seconds.", logfile, 1)

    return latent


def trainaae(
    data_loader: DataLoader,
    aae_options: AAEOptions,
    training_options: AAETrainingOptions,
    vamb_options: VambOptions,
    lrate: float,
    alpha: Optional[float],  # set automatically if None
    logfile: IO[str],
    contignames: Sequence[str],
) -> tuple[np.ndarray, dict[str, set[str]]]:
    begintime = time.time()
    log("\nCreating and training AAE", logfile)
    nsamples = data_loader.dataset.tensors[0].shape[1]

    aae = vamb.aamb_encode.AAE(
        nsamples,
        aae_options.nhiddens,
        aae_options.nlatent_z,
        aae_options.nlatent_y,
        aae_options.sl,
        aae_options.slr,
        alpha,
        _cuda=vamb_options.cuda,
        seed=vamb_options.seed,
    )

    log("Created AAE", logfile, 1)
    modelpath = os.path.join(vamb_options.out_dir, "aae_model.pt")
    aae.trainmodel(
        vamb.encode.set_batchsize(data_loader, training_options.batchsize),
        training_options.nepochs,
        training_options.batchsteps,
        training_options.temp,
        lrate,
        logfile,
        modelpath,
    )

    print("", file=logfile)
    log("Encoding to latent representation", logfile, 1)
    clusters_y_dict, latent = aae.get_latents(contignames, data_loader)
    vamb.vambtools.write_npz(
        os.path.join(vamb_options.out_dir, "aae_z_latent.npz"), latent
    )

    del aae  # Needed to free "latent" array's memory references?

    elapsed = round(time.time() - begintime, 2)
    log(f"Trained AAE and encoded in {elapsed} seconds.", logfile, 1)

    return latent, clusters_y_dict


def cluster(
    cluster_options: ClusterOptions,
    clusterspath: Path,
    latent: np.ndarray,
    contignames: Sequence[str],  # of dtype object
    vamb_options: VambOptions,
    logfile: IO[str],
    cluster_prefix: str,
) -> None:
    begintime = time.time()

    log("\nClustering", logfile)
    log(f"Windowsize: {cluster_options.window_size}", logfile, 1)
    log(
        f"Min successful thresholds detected: {cluster_options.min_successes}",
        logfile,
        1,
    )
    log(f"Max clusters: {cluster_options.max_clusters}", logfile, 1)
    log(f"Min cluster size: {cluster_options.min_cluster_size}", logfile, 1)
    log(f"Use CUDA for clustering: {vamb_options.cuda}", logfile, 1)
    log(
        "Separator: {}".format(
            None
            if cluster_options.binsplit_separator is None
            else ('"' + cluster_options.binsplit_separator + '"')
        ),
        logfile,
        1,
    )

    cluster_generator = vamb.cluster.ClusterGenerator(
        latent,
        windowsize=cluster_options.window_size,
        minsuccesses=cluster_options.min_successes,
        destroy=True,
        normalized=False,
        cuda=vamb_options.cuda,
        seed=vamb_options.seed,
    )

    renamed = (
        (str(cluster_index + 1), {contignames[i] for i in members})
        for (cluster_index, (_, members)) in enumerate(
            map(lambda x: x.as_tuple(), cluster_generator)
        )
    )

    # Binsplit if given a separator
    if cluster_options.binsplit_separator is not None:
        maybe_split = vamb.vambtools.binsplit(
            renamed, cluster_options.binsplit_separator
        )
    else:
        maybe_split = renamed

    with open(clusterspath, "w") as clustersfile:
        clusternumber, ncontigs = vamb.vambtools.write_clusters(
            clustersfile,
            maybe_split,
            max_clusters=cluster_options.max_clusters,
            min_size=cluster_options.min_cluster_size,
            rename=False,
            cluster_prefix=cluster_prefix,
        )

    print("", file=logfile)
    log(f"Clustered {ncontigs} contigs in {clusternumber} bins", logfile, 1)

    elapsed = round(time.time() - begintime, 2)
    log(f"Clustered contigs in {elapsed} seconds.", logfile, 1)


def write_fasta(
    outdir: Path,
    clusterspath: Path,
    fastapath: Path,
    contignames: Sequence[str],  # of object
    contiglengths: np.ndarray,
    minfasta: int,
    logfile: IO[str],
    separator: Optional[str],
) -> None:
    begintime = time.time()

    log("\nWriting FASTA files", logfile)
    log("Minimum FASTA size: " + str(minfasta), logfile, 1)
    assert len(contignames) == len(contiglengths)

    lengthof = dict(zip(contignames, contiglengths))
    filtered_clusters: dict[str, set[str]] = dict()

    with open(clusterspath) as file:
        clusters = vamb.vambtools.read_clusters(file)

    for cluster, contigs in clusters.items():
        size = sum(lengthof[contig] for contig in contigs)
        if size >= minfasta:
            filtered_clusters[cluster] = clusters[cluster]

    del lengthof, clusters
    keep: set[str] = set()
    for contigs in filtered_clusters.values():
        keep.update(set(contigs))

    with vamb.vambtools.Reader(fastapath) as file:
        vamb.vambtools.write_bins(
            outdir.joinpath("bins"),
            filtered_clusters,
            file,
            maxbins=None,
            separator=separator,
        )

    ncontigs = sum(map(len, filtered_clusters.values()))
    nfiles = len(filtered_clusters)
    print("", file=logfile)
    log(f"Wrote {ncontigs} contigs to {nfiles} FASTA files", logfile, 1)

    elapsed = round(time.time() - begintime, 2)
    log(f"Wrote FASTA in {elapsed} seconds.", logfile, 1)


def run(
    vamb_options: VambOptions,
    comp_options: CompositionOptions,
    abundance_options: AbundanceOptions,
    encoder_options: EncoderOptions,
    training_options: TrainingOptions,
    cluster_options: ClusterOptions,
    logfile: IO[str],
):
    vae_options = encoder_options.vae_options
    aae_options = encoder_options.aae_options
    vae_training_options = training_options.vae_options
    aae_training_options = training_options.aae_options

    log("Starting Vamb version " + ".".join(map(str, vamb.__version__)), logfile)
    log("Date and time is " + str(datetime.datetime.now()), logfile, 1)
    begintime = time.time()

    # Get TNFs, save as npz
    composition = calc_tnf(comp_options, vamb_options.out_dir, logfile)

    # Parse BAMs, save as npz
    abundance = calc_rpkm(
        abundance_options,
        vamb_options.out_dir,
        composition.metadata,
        vamb_options.n_threads,
        logfile,
    )
    time_generating_input = round(time.time() - begintime, 2)
    log(
        f"\nTNF and coabundances generated in {time_generating_input} seconds.",
        logfile,
        1,
    )

    data_loader, mask = vamb.encode.make_dataloader(
        abundance.matrix,
        composition.matrix,
        composition.metadata.lengths,
        256,  # dummy value - we change this before using the actual loader
        destroy=True,
        cuda=vamb_options.cuda,
    )
    composition.metadata.filter_mask(mask)

    print("", file=logfile)
    log("Created dataloader and mask", logfile, 0)
    vamb.vambtools.write_npz(vamb_options.out_dir.joinpath("mask.npz"), mask)
    n_discarded = len(mask) - mask.sum()
    log(f"Number of sequences unsuitable for encoding: {n_discarded}", logfile, 1)
    log(f"Number of sequences remaining: {len(mask) - n_discarded}", logfile, 1)

    latent = None
    if vae_options is not None:
        assert vae_training_options is not None
        # Train, save model
        latent = trainvae(
            vae_options=vae_options,
            training_options=vae_training_options,
            vamb_options=vamb_options,
            lrate=training_options.lrate,
            alpha=encoder_options.alpha,
            data_loader=data_loader,
            logfile=logfile,
        )
        print("", file=logfile)

    clusters_y_dict = None
    latent_z = None
    if aae_options is not None:
        assert aae_training_options is not None
        # Train, save model
        latent_z, clusters_y_dict = trainaae(
            data_loader=data_loader,
            aae_options=aae_options,
            vamb_options=vamb_options,
            training_options=aae_training_options,
            lrate=training_options.lrate,
            alpha=encoder_options.alpha,
            logfile=logfile,
            contignames=composition.metadata.identifiers,
        )
        print("", file=logfile)

    # Free up memory
    comp_metadata = composition.metadata
    del composition, abundance

    # Write contignames and contiglengths needed for dereplication purposes
    np.savetxt(
        vamb_options.out_dir.joinpath("contignames"),
        comp_metadata.identifiers,
        fmt="%s",
    )
    np.savez(vamb_options.out_dir.joinpath("lengths.npz"), comp_metadata.lengths)

    if vae_options is not None:
        assert latent is not None
        assert comp_metadata.nseqs == len(latent)
        # Cluster, save tsv file
        clusterspath = vamb_options.out_dir.joinpath("vae_clusters.tsv")
        cluster(
            cluster_options,
            clusterspath,
            latent,
            comp_metadata.identifiers,
            vamb_options,
            logfile,
            "vae_",
        )
        log("VAE latent clustered", logfile, 1)

        del latent
        fin_cluster_latent = time.time()

        if vamb_options.min_fasta_output_size is not None:
            path = comp_options.path
            assert isinstance(path, FASTAPath)
            write_fasta(
                vamb_options.out_dir,
                clusterspath,
                path,
                comp_metadata.identifiers,
                comp_metadata.lengths,
                vamb_options.min_fasta_output_size,
                logfile,
                separator=cluster_options.binsplit_separator,
            )

        writing_bins_time = round(time.time() - fin_cluster_latent, 2)
        log(f"VAE bins written in {writing_bins_time} seconds.", logfile, 1)

    if aae_options is not None:
        assert latent_z is not None
        assert clusters_y_dict is not None
        assert comp_metadata.nseqs == len(latent_z)
        # Cluster, save tsv file
        clusterspath = vamb_options.out_dir.joinpath("aae_z_clusters.tsv")

        cluster(
            cluster_options,
            clusterspath,
            latent_z,
            comp_metadata.identifiers,
            vamb_options,
            logfile,
            "aae_z_",
        )

        fin_cluster_latent_z = time.time()
        log("AAE z latent clustered.", logfile, 1)

        del latent_z

        if vamb_options.min_fasta_output_size is not None:
            path = comp_options.path
            assert isinstance(path, FASTAPath)
            write_fasta(
                vamb_options.out_dir,
                clusterspath,
                path,
                comp_metadata.identifiers,
                comp_metadata.lengths,
                vamb_options.min_fasta_output_size,
                logfile,
                separator=cluster_options.binsplit_separator,
            )
        time_writing_bins_z = time.time()
        writing_bins_time_z = round(time_writing_bins_z - fin_cluster_latent_z, 2)
        log(f"AAE z bins written in {writing_bins_time_z} seconds.", logfile, 1)

        clusterspath = vamb_options.out_dir.joinpath("aae_y_clusters.tsv")
        # Binsplit if given a separator
        if cluster_options.binsplit_separator is not None:
            maybe_split = vamb.vambtools.binsplit(
                clusters_y_dict.items(), cluster_options.binsplit_separator
            )
        else:
            maybe_split = clusters_y_dict.items()
        with open(clusterspath, "w") as clustersfile:
            clusternumber, ncontigs = vamb.vambtools.write_clusters(
                clustersfile,
                maybe_split,
                max_clusters=cluster_options.max_clusters,
                min_size=cluster_options.min_cluster_size,
                rename=False,
                cluster_prefix="aae_y_",
            )

        print("", file=logfile)
        log(f"Clustered {ncontigs} contigs in {clusternumber} bins", logfile, 1)
        time_start_writin_z_bins = time.time()
        if vamb_options.min_fasta_output_size is not None:
            path = comp_options.path
            assert isinstance(path, FASTAPath)
            write_fasta(
                vamb_options.out_dir,
                clusterspath,
                path,
                comp_metadata.identifiers,
                comp_metadata.lengths,
                vamb_options.min_fasta_output_size,
                logfile,
                separator=cluster_options.binsplit_separator,
            )
        time_writing_bins_y = time.time()
        writing_bins_time_y = round(time_writing_bins_y - time_start_writin_z_bins, 2)
        log(f"AAE y bins written in {writing_bins_time_y} seconds.", logfile, 1)

    log(f"\nCompleted Vamb in {round(time.time() - begintime, 2)} seconds.", logfile, 0)


def main():
    doc = f"""Avamb: Adversarial and Variational autoencoders for metagenomic binning.

    Version: {'.'.join([str(i) for i in vamb.__version__])}

    Default use, good for most datasets:
    vamb --outdir out --fasta my_contigs.fna --bamfiles *.bam -o C

    For advanced use and extensions of Avamb, check documentation of the package
    at https://github.com/RasmussenLab/vamb."""  # must be updated with the new github
    parser = argparse.ArgumentParser(
        prog="vamb",
        description=doc,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        usage="%(prog)s outdir tnf_input rpkm_input [options]",
        add_help=False,
    )

    # Help
    helpos = parser.add_argument_group(title="Help and version", description=None)
    helpos.add_argument("-h", "--help", help="print help and exit", action="help")
    helpos.add_argument(
        "--version",
        action="version",
        version=f'Vamb {".".join(map(str, vamb.__version__))}',
    )

    # Positional arguments
    reqos = parser.add_argument_group(title="Output (required)", description=None)
    reqos.add_argument(
        "--outdir",
        metavar="",
        type=Path,
        required=True,
        help="output directory to create",
    )

    # Other arguments
    otheros = parser.add_argument_group(title="Other options", description=None)
    otheros.add_argument(
        "--seed",
        metavar="",
        type=int,
        default=0,
        help="Random seed [random] (determinism not guaranteed)",
    )

    # TNF arguments
    tnfos = parser.add_argument_group(
        title="TNF input (either fasta or all .npz files required)"
    )
    tnfos.add_argument("--fasta", metavar="", type=Path, help="path to fasta file")
    tnfos.add_argument(
        "--composition", metavar="", type=Path, help="path to .npz of composition"
    )

    # RPKM arguments
    rpkmos = parser.add_argument_group(
        title="RPKM input (either BAMs or .npz required)"
    )
    rpkmos.add_argument(
        "--bamfiles",
        dest="bampaths",
        metavar="",
        type=Path,
        help="paths to (multiple) BAM files",
        nargs="+",
    )
    rpkmos.add_argument(
        "--rpkm",
        metavar="",
        dest="abundancepath",
        type=Path,
        help="path to .npz of RPKM (abundances)",
    )

    # Optional arguments
    inputos = parser.add_argument_group(title="IO options", description=None)

    inputos.add_argument(
        "-m",
        dest="minlength",
        metavar="",
        type=int,
        default=250,
        help="ignore contigs shorter than this [250]",
    )
    inputos.add_argument(
        "-z",
        dest="min_alignment_id",
        metavar="",
        type=float,
        default=None,
        help="ignore reads with nucleotide identity below this [0.0]",
    )
    inputos.add_argument(
        "-p",
        dest="nthreads",
        metavar="",
        type=int,
        default=DEFAULT_THREADS,
        help=(
            "number of threads to use " "[min(" + str(DEFAULT_THREADS) + ", nbamfiles)]"
        ),
    )
    inputos.add_argument(
        "--norefcheck",
        help="skip reference name hashing check [False]",
        action="store_true",
    )
    inputos.add_argument(
        "--minfasta",
        dest="min_fasta_output_size",
        metavar="",
        type=int,
        default=None,
        help="minimum bin size to output as fasta [None = no files]",
    )

    # Model selection argument
    model_selection = parser.add_argument_group(
        title="Model selection", description=None
    )

    model_selection.add_argument(
        "--model",
        dest="model",
        metavar="",
        type=str,
        choices=["vae", "aae", "vae-aae"],
        default="vae",
        help="Choose which model to run; only vae (vae), only aae (aae), the combination of vae and aae (vae-aae), [vae]",
    )

    # VAE arguments
    vaeos = parser.add_argument_group(title="VAE options", description=None)

    vaeos.add_argument(
        "-n",
        dest="nhiddens",
        metavar="",
        type=int,
        nargs="+",
        default=None,
        help="hidden neurons [Auto]",
    )
    vaeos.add_argument(
        "-l",
        dest="nlatent",
        metavar="",
        type=int,
        default=32,
        help="latent neurons [32]",
    )
    vaeos.add_argument(
        "-a",
        dest="alpha",
        metavar="",
        type=float,
        default=None,
        help="alpha, weight of TNF versus depth loss [Auto]",
    )
    vaeos.add_argument(
        "-b",
        dest="beta",
        metavar="",
        type=float,
        default=200.0,
        help="beta, capacity to learn [200.0]",
    )
    vaeos.add_argument(
        "-d",
        dest="dropout",
        metavar="",
        type=float,
        default=None,
        help="dropout [Auto]",
    )
    vaeos.add_argument(
        "--cuda", help="use GPU to train & cluster [False]", action="store_true"
    )

    trainos = parser.add_argument_group(title="Training options", description=None)

    trainos.add_argument(
        "-e", dest="nepochs", metavar="", type=int, default=300, help="epochs [300]"
    )
    trainos.add_argument(
        "-t",
        dest="batchsize",
        metavar="",
        type=int,
        default=256,
        help="starting batch size [256]",
    )
    trainos.add_argument(
        "-q",
        dest="batchsteps",
        metavar="",
        type=int,
        nargs="*",
        default=[25, 75, 150, 225],
        help="double batch size at epochs [25 75 150 225]",
    )
    trainos.add_argument(
        "-r",
        dest="lrate",
        metavar="",
        type=float,
        default=1e-3,
        help="learning rate [0.001]",
    )
    # AAE arguments
    aaeos = parser.add_argument_group(title="AAE options", description=None)

    aaeos.add_argument(
        "--n_aae",
        dest="nhiddens_aae",
        metavar="",
        type=int,
        default=547,
        help="hidden neurons AAE [547]",
    )
    aaeos.add_argument(
        "--z_aae",
        dest="nlatent_aae_z",
        metavar="",
        type=int,
        default=283,
        help="latent neurons AAE continuous latent space  [283]",
    )
    aaeos.add_argument(
        "--y_aae",
        dest="nlatent_aae_y",
        metavar="",
        type=int,
        default=700,
        help="latent neurons AAE categorical latent space [700]",
    )
    aaeos.add_argument(
        "--sl_aae",
        dest="sl",
        metavar="",
        type=float,
        default=0.00964,
        help="loss scale between reconstruction loss and adversarial losses [0.00964] ",
    )
    aaeos.add_argument(
        "--slr_aae",
        dest="slr",
        metavar="",
        type=float,
        default=0.5,
        help="loss scale between reconstruction adversarial losses [0.5] ",
    )
    aaeos.add_argument(
        "--aae_temp",
        dest="temp",
        metavar="",
        type=float,
        default=0.1596,
        help=" Temperature of the softcategorical prior [0.1596]",
    )

    aaetrainos = parser.add_argument_group(
        title="Training options AAE", description=None
    )

    aaetrainos.add_argument(
        "--e_aae",
        dest="nepochs_aae",
        metavar="",
        type=int,
        default=70,
        help="epochs AAE [70]",
    )
    aaetrainos.add_argument(
        "--t_aae",
        dest="batchsize_aae",
        metavar="",
        type=int,
        default=256,
        help="starting batch size AAE [256]",
    )
    aaetrainos.add_argument(
        "--q_aae",
        dest="batchsteps_aae",
        metavar="",
        type=int,
        nargs="*",
        default=[25, 50],
        help="double batch size at epochs AAE [25,50]",
    )
    # Clustering arguments
    clusto = parser.add_argument_group(title="Clustering options", description=None)
    clusto.add_argument(
        "-w",
        dest="window_size",
        metavar="",
        type=int,
        default=200,
        help="size of window to count successes [200]",
    )
    clusto.add_argument(
        "-u",
        dest="min_successes",
        metavar="",
        type=int,
        default=20,
        help="minimum success in window [20]",
    )
    clusto.add_argument(
        "-i",
        dest="min_cluster_size",
        metavar="",
        type=int,
        default=1,
        help="minimum cluster size [1]",
    )
    clusto.add_argument(
        "-c",
        dest="max_clusters",
        metavar="",
        type=int,
        default=None,
        help="stop after c clusters [None = infinite]",
    )
    clusto.add_argument(
        "-o",
        dest="binsplit_separator",
        metavar="",
        type=str,
        default=None,
        help="binsplit separator [None = no split]",
    )

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit()

    args = parser.parse_args()

    comp_options = CompositionOptions(args.fasta, args.composition, args.minlength)

    abundance_options = AbundanceOptions(
        args.bampaths,
        args.abundancepath,
        args.min_alignment_id,
        not args.norefcheck,
    )

    if args.model in ("vae", "vae-aae"):
        vae_options = VAEOptions(
            nhiddens=args.nhiddens,
            nlatent=args.nlatent,
            beta=args.beta,
            dropout=args.dropout,
        )

        vae_training_options = VAETrainingOptions(
            nepochs=args.nepochs, batchsize=args.batchsize, batchsteps=args.batchsteps
        )
    else:
        assert args.model == "aae"
        vae_options = None
        vae_training_options = None

        for name, arg, default in [
            ("-n", args.nhiddens, None),
            ("-l", args.nlatent, 32),
            ("-b", args.beta, 200.0),
            ("-d", args.dropout, None),
            ("-t", args.batchsize, 256),
            ("-q", args.batchsteps, [25, 75, 150, 225]),
        ]:
            if arg != default:
                raise ValueError(
                    f'VAE model not used, but VAE-specific arg "{name}" used'
                )

    if args.model in ("aae", "vae-aae"):
        aae_options = AAEOptions(
            nhiddens=args.nhiddens_aae,
            nlatent_z=args.nlatent_aae_z,
            nlatent_y=args.nlatent_aae_y,
            sl=args.sl,
            slr=args.slr,
        )

        aae_training_options = AAETrainingOptions(
            nepochs=args.nepochs_aae,
            batchsize=args.batchsize_aae,
            batchsteps=args.batchsteps_aae,
            temp=args.temp,
        )
    else:
        assert args.model == "vae"
        aae_options = None
        aae_training_options = None

        for name, arg, default in [
            ("--n_aae", args.nhiddens_aae, 547),
            ("--z_aae", args.nlatent_aae_z, 283),
            ("--y_aae", args.nlatent_aae_y, 700),
            ("--sl_aae", args.sl, 0.00964),
            ("--slr_aae", args.slr, 0.5),
            ("--aae_temp", args.temp, 0.1596),
            ("--e_aae", args.nepochs_aae, 70),
            ("--t_aae", args.batchsize_aae, 256),
            ("--q_aae", args.batchsteps_aae, [25, 50]),
        ]:
            if arg != default:
                raise ValueError(
                    f'AAE model not used, but AAE-specific arg "{name}" used'
                )

    encoder_options = EncoderOptions(
        vae_options=vae_options, aae_options=aae_options, alpha=args.alpha
    )
    training_options = TrainingOptions(
        encoder_options=encoder_options,
        vae_options=vae_training_options,
        aae_options=aae_training_options,
        lrate=args.lrate,
    )

    cluster_options = ClusterOptions(
        args.window_size,
        args.min_successes,
        args.min_cluster_size,
        args.max_clusters,
        args.binsplit_separator,
    )

    vamb_options = VambOptions(
        args.outdir,
        args.nthreads,
        comp_options,
        args.min_fasta_output_size,
        args.seed,
        args.cuda,
    )

    torch.set_num_threads(vamb_options.n_threads)

    try_make_dir(vamb_options.out_dir)

    with open(vamb_options.out_dir.joinpath("log.txt"), "w") as logfile:
        run(
            vamb_options=vamb_options,
            comp_options=comp_options,
            abundance_options=abundance_options,
            encoder_options=encoder_options,
            training_options=training_options,
            cluster_options=cluster_options,
            logfile=logfile,
        )


if __name__ == "__main__":
    main()



import unittest
import numpy as np
from hashlib import md5
import random
import string

import vamb


class TestClusterer(unittest.TestCase):
    data = np.random.random((1024, 40)).astype(np.float32)

    def test_bad_params(self):
        with self.assertRaises(ValueError):
            vamb.cluster.ClusterGenerator(self.data.astype(np.float64))

        with self.assertRaises(ValueError):
            vamb.cluster.ClusterGenerator(self.data, maxsteps=0)

        with self.assertRaises(ValueError):
            vamb.cluster.ClusterGenerator(self.data, windowsize=0)

        with self.assertRaises(ValueError):
            vamb.cluster.ClusterGenerator(self.data, minsuccesses=0)

        with self.assertRaises(ValueError):
            vamb.cluster.ClusterGenerator(self.data, minsuccesses=5, windowsize=4)

        with self.assertRaises(ValueError):
            vamb.cluster.ClusterGenerator(np.random.random((0, 40)))

    # In the code, in __init__ of the cluster generator, the input matrix
    # is shuffled, and an index array is permuted to keep track of which
    # indices was which.
    # This depends on the implementation of shuffling and permute being
    # the same, which I test here.
    def test_shuffling(self):
        cp = self.data.copy()
        np.random.RandomState(0).shuffle(cp)
        indices = np.random.RandomState(0).permutation(len(cp))
        cplike = self.data[indices]
        self.assertTrue(np.all(cplike == cp))

    def test_basics(self):
        clstr = vamb.cluster.ClusterGenerator(self.data)
        self.assertIs(clstr, iter(clstr))

        x = next(clstr)
        self.assertIsInstance(x, vamb.cluster.Cluster)

        clusters = list(clstr)
        clusters.append(x)

        # All members are clustered
        self.assertEqual(sum(map(lambda x: len(x.members), clusters)), len(self.data))

        # Elements of members are exactly the matrix row indices
        mems = set()
        for i in clusters:
            mems.update(i.members)
        self.assertEqual(mems, set(range(len(self.data))))

    def test_detruction(self):
        copy = self.data.copy()
        clstr = vamb.cluster.ClusterGenerator(self.data)
        self.assertTrue(np.any(np.abs(self.data - clstr.matrix.numpy()) > 0.001))
        clstr = vamb.cluster.ClusterGenerator(copy, destroy=True)
        self.assertTrue(np.all(np.abs(copy - clstr.matrix.numpy()) < 1e-6))
        self.assertTrue(np.any(np.abs(self.data - clstr.matrix.numpy()) > 0.001))

    @staticmethod
    def xor_rows_hash(matrix):
        m = np.frombuffer(matrix.copy().data, dtype=np.uint32)
        m.shape = matrix.shape
        v = m[0]
        for i in range(1, len(m)):
            v ^= m[i]
        return md5(v).digest().hex()

    def test_normalization(self):
        hash_before = md5(self.data.data.tobytes()).digest().hex()
        vamb.cluster.ClusterGenerator(self.data)
        self.assertEqual(hash_before, md5(self.data.data.tobytes()).digest().hex())
        cp = self.data.copy()
        vamb.cluster.ClusterGenerator(cp, destroy=True)
        hash_after = md5(cp.data.tobytes()).digest().hex()
        self.assertNotEqual(hash_before, hash_after)

        # Rows are permuted by the clusterer. We use xor to check the rows
        # are still essentially the same.
        before_xor = self.xor_rows_hash(cp)
        vamb.cluster.ClusterGenerator(cp, destroy=True, normalized=True)
        self.assertEqual(before_xor, self.xor_rows_hash(cp))

    def test_cluster(self):
        x = next(vamb.cluster.ClusterGenerator(self.data))
        self.assertIsInstance(x.members, np.ndarray)
        med, st = x.as_tuple()
        self.assertIsInstance(st, set)
        self.assertEqual(set(x.members), st)
        self.assertIn(med, st)


class TestPairs(unittest.TestCase):
    n_samples = 1024
    data = np.random.random((n_samples, 40)).astype(np.float32)

    @staticmethod
    def randstring(len):
        return "".join(random.choices(string.ascii_letters, k=len))

    def test_too_few_names(self):
        clstr = vamb.cluster.ClusterGenerator(self.data)
        nameset = {self.randstring(10) for i in range(len(self.data) - 1)}
        with self.assertRaises(ValueError):
            list(vamb.cluster.pairs(clstr, list(nameset)))

    def test_pairs(self):
        clstr = vamb.cluster.ClusterGenerator(self.data)
        nameset = {self.randstring(10) for i in range(len(self.data))}
        pairs = list(vamb.cluster.pairs(clstr, list(nameset)))

        medoid, members = pairs[0]
        self.assertIsInstance(medoid, str)
        self.assertIsInstance(members, set)
        self.assertTrue(all(len(i[1]) > 0 for i in pairs))
        self.assertTrue(sum(map(lambda i: len(i[1]), pairs)), len(nameset))
        allmembers = set()
        for medoid, mems in pairs:
            allmembers.update(mems)
        self.assertEqual(allmembers, nameset)



from typing import Optional, IO, Union
from pathlib import Path
import vamb.vambtools as _vambtools
from torch.utils.data.dataset import TensorDataset as _TensorDataset
from torch.utils.data import DataLoader as _DataLoader
from torch.nn.functional import softmax as _softmax
from torch.optim import Adam as _Adam
from torch import Tensor
from torch import nn as _nn
from math import log as _log

__doc__ = """Encode a depths matrix and a tnf matrix to latent representation.

Creates a variational autoencoder in PyTorch and tries to represent the depths
and tnf in the latent space under gaussian noise.

Usage:
>>> vae = VAE(nsamples=6)
>>> dataloader, mask = make_dataloader(depths, tnf, lengths)
>>> vae.trainmodel(dataloader)
>>> latent = vae.encode(dataloader) # Encode to latent representation
>>> latent.shape
(183882, 32)
"""

__cmd_doc__ = """Encode depths and TNF using a VAE to latent representation"""

import numpy as _np
import torch as _torch


def set_batchsize(
    data_loader: _DataLoader, batch_size: int, encode=False
) -> _DataLoader:
    """Effectively copy the data loader, but with a different batch size.

    The `encode` option is used to copy the dataloader to use before encoding.
    This will not drop the last minibatch whose size may not be the requested
    batch size, and will also not shuffle the data.
    """
    return _DataLoader(
        dataset=data_loader.dataset,
        batch_size=batch_size,
        shuffle=not encode,
        drop_last=not encode,
        num_workers=1 if encode else data_loader.num_workers,
        pin_memory=data_loader.pin_memory,
    )


def make_dataloader(
    rpkm: _np.ndarray,
    tnf: _np.ndarray,
    lengths: _np.ndarray,
    batchsize: int = 256,
    destroy: bool = False,
    cuda: bool = False,
) -> tuple[_DataLoader[tuple[Tensor, Tensor, Tensor]], _np.ndarray]:
    """Create a DataLoader and a contig mask from RPKM and TNF.

    The dataloader is an object feeding minibatches of contigs to the VAE.
    The data are normalized versions of the input datasets, with zero-contigs,
    i.e. contigs where a row in either TNF or RPKM are all zeros, removed.
    The mask is a boolean mask designating which contigs have been kept.

    Inputs:
        rpkm: RPKM matrix (N_contigs x N_samples)
        tnf: TNF matrix (N_contigs x N_TNF)
        batchsize: Starting size of minibatches for dataloader
        destroy: Mutate rpkm and tnf array in-place instead of making a copy.
        cuda: Pagelock memory of dataloader (use when using GPU acceleration)

    Outputs:
        DataLoader: An object feeding data to the VAE
        mask: A boolean mask of which contigs are kept
    """

    if not isinstance(rpkm, _np.ndarray) or not isinstance(tnf, _np.ndarray):
        raise ValueError("TNF and RPKM must be Numpy arrays")

    if batchsize < 1:
        raise ValueError(f"Batch size must be minimum 1, not {batchsize}")

    if len(rpkm) != len(tnf) or len(tnf) != len(lengths):
        raise ValueError("Lengths of RPKM, TNF and lengths arrays must be the same")

    if not (rpkm.dtype == tnf.dtype == _np.float32):
        raise ValueError("TNF and RPKM must be Numpy arrays of dtype float32")

    ### Copy arrays and mask them ###
    # Copy if not destroy - this way we can have all following operations in-place
    # for simplicity
    if not destroy:
        rpkm = rpkm.copy()
        tnf = tnf.copy()

    # Normalize samples to have same depth
    sample_depths_sum = rpkm.sum(axis=0)
    if _np.any(sample_depths_sum == 0):
        raise ValueError(
            "One or more samples have zero depth in all sequences, so cannot be depth normalized"
        )
    rpkm *= 1_000_000 / sample_depths_sum

    # If multiple samples, also include nonzero depth as requirement for accept
    # of sequences
    mask = tnf.sum(axis=1) != 0
    depthssum = None
    if rpkm.shape[1] > 1:
        depthssum = rpkm.sum(axis=1)
        mask &= depthssum != 0
        depthssum = depthssum[mask]
        assert isinstance(depthssum, _np.ndarray)

    if mask.sum() < batchsize:
        raise ValueError(
            "Fewer sequences left after filtering than the batch size. "
            + "This probably means you try to run on a too small dataset (below ~10k sequences), "
            + "or that nearly all sequences were filtered away. Check the log file, "
            + "and verify BAM file content is sensible."
        )

    _vambtools.numpy_inplace_maskarray(rpkm, mask)
    _vambtools.numpy_inplace_maskarray(tnf, mask)

    # If multiple samples, normalize to sum to 1, else zscore normalize
    if rpkm.shape[1] > 1:
        assert depthssum is not None  # we set it so just above
        rpkm /= depthssum.reshape((-1, 1))
    else:
        _vambtools.zscore(rpkm, axis=0, inplace=True)

    # Normalize TNF
    _vambtools.zscore(tnf, axis=0, inplace=True)

    # Create weights
    lengths = (lengths[mask]).astype(_np.float32)
    weights = _np.log(lengths).astype(_np.float32) - 5.0
    weights[weights < 2.0] = 2.0
    weights *= len(weights) / weights.sum()
    weights.shape = (len(weights), 1)

    ### Create final tensors and dataloader ###
    depthstensor = _torch.from_numpy(rpkm)  # this is a no-copy operation
    tnftensor = _torch.from_numpy(tnf)
    weightstensor = _torch.from_numpy(weights)
    n_workers = 4 if cuda else 1
    dataset = _TensorDataset(depthstensor, tnftensor, weightstensor)
    dataloader = _DataLoader(
        dataset=dataset,
        batch_size=batchsize,
        drop_last=True,
        shuffle=True,
        num_workers=n_workers,
        pin_memory=cuda,
    )

    return dataloader, mask


class VAE(_nn.Module):
    """Variational autoencoder, subclass of torch.nn.Module.

    Instantiate with:
        nsamples: Number of samples in abundance matrix
        nhiddens: list of n_neurons in the hidden layers [None=Auto]
        nlatent: Number of neurons in the latent layer [32]
        alpha: Approximate starting TNF/(CE+TNF) ratio in loss. [None = Auto]
        beta: Multiply KLD by the inverse of this value [200]
        dropout: Probability of dropout on forward pass [0.2]
        cuda: Use CUDA (GPU accelerated training) [False]

    vae.trainmodel(dataloader, nepochs batchsteps, lrate, logfile, modelfile)
        Trains the model, returning None

    vae.encode(self, data_loader):
        Encodes the data in the data loader and returns the encoded matrix.

    If alpha or dropout is None and there is only one sample, they are set to
    0.99 and 0.0, respectively
    """

    def __init__(
        self,
        nsamples: int,
        nhiddens: Optional[list[int]] = None,
        nlatent: int = 32,
        alpha: Optional[float] = None,
        beta: float = 200.0,
        dropout: Optional[float] = 0.2,
        cuda: bool = False,
        seed: int = 0,
    ):
        if nlatent < 1:
            raise ValueError(f"Minimum 1 latent neuron, not {nlatent}")

        if nsamples < 1:
            raise ValueError(f"nsamples must be > 0, not {nsamples}")

        # If only 1 sample, we weigh alpha and nhiddens differently
        if alpha is None:
            alpha = 0.15 if nsamples > 1 else 0.50

        if nhiddens is None:
            nhiddens = [512, 512] if nsamples > 1 else [256, 256]

        if dropout is None:
            dropout = 0.2 if nsamples > 1 else 0.0

        if any(i < 1 for i in nhiddens):
            raise ValueError(f"Minimum 1 neuron per layer, not {min(nhiddens)}")

        if beta <= 0:
            raise ValueError(f"beta must be > 0, not {beta}")

        if not (0 < alpha < 1):
            raise ValueError(f"alpha must be 0 < alpha < 1, not {alpha}")

        if not (0 <= dropout < 1):
            raise ValueError(f"dropout must be 0 <= dropout < 1, not {dropout}")

        _torch.manual_seed(seed)
        super(VAE, self).__init__()

        # Initialize simple attributes
        self.usecuda = cuda
        self.nsamples = nsamples
        self.ntnf = 103
        self.alpha = alpha
        self.beta = beta
        self.nhiddens = nhiddens
        self.nlatent = nlatent
        self.dropout = dropout

        # Initialize lists for holding hidden layers
        self.encoderlayers = _nn.ModuleList()
        self.encodernorms = _nn.ModuleList()
        self.decoderlayers = _nn.ModuleList()
        self.decodernorms = _nn.ModuleList()

        # Add all other hidden layers
        for nin, nout in zip(
            [self.nsamples + self.ntnf] + self.nhiddens, self.nhiddens
        ):
            self.encoderlayers.append(_nn.Linear(nin, nout))
            self.encodernorms.append(_nn.BatchNorm1d(nout))

        # Latent layers
        self.mu = _nn.Linear(self.nhiddens[-1], self.nlatent)
        self.logsigma = _nn.Linear(self.nhiddens[-1], self.nlatent)

        # Add first decoding layer
        for nin, nout in zip([self.nlatent] + self.nhiddens[::-1], self.nhiddens[::-1]):
            self.decoderlayers.append(_nn.Linear(nin, nout))
            self.decodernorms.append(_nn.BatchNorm1d(nout))

        # Reconstruction (output) layer
        self.outputlayer = _nn.Linear(self.nhiddens[0], self.nsamples + self.ntnf)

        # Activation functions
        self.relu = _nn.LeakyReLU()
        self.softplus = _nn.Softplus()
        self.dropoutlayer = _nn.Dropout(p=self.dropout)

        if cuda:
            self.cuda()

    def _encode(self, tensor: Tensor) -> tuple[Tensor, Tensor]:
        tensors = list()

        # Hidden layers
        for encoderlayer, encodernorm in zip(self.encoderlayers, self.encodernorms):
            tensor = encodernorm(self.dropoutlayer(self.relu(encoderlayer(tensor))))
            tensors.append(tensor)

        # Latent layers
        mu = self.mu(tensor)

        # Note: This softplus constrains logsigma to positive. As reconstruction loss pushes
        # logsigma as low as possible, and KLD pushes it towards 0, the optimizer will
        # always push this to 0, meaning that the logsigma layer will be pushed towards
        # negative infinity. This creates a nasty numerical instability in VAMB. Luckily,
        # the gradient also disappears as it decreases towards negative infinity, avoiding
        # NaN poisoning in most cases. We tried to remove the softplus layer, but this
        # necessitates a new round of hyperparameter optimization, and there is no way in
        # hell I am going to do that at the moment of writing.
        # Also remove needless factor 2 in definition of latent in reparameterize function.
        logsigma = self.softplus(self.logsigma(tensor))

        return mu, logsigma

    # sample with gaussian noise
    def reparameterize(self, mu: Tensor, logsigma: Tensor) -> Tensor:
        epsilon = _torch.randn(mu.size(0), mu.size(1))

        if self.usecuda:
            epsilon = epsilon.cuda()

        epsilon.requires_grad = True

        # See comment above regarding softplus
        latent = mu + epsilon * _torch.exp(logsigma / 2)

        return latent

    def _decode(self, tensor: Tensor) -> tuple[Tensor, Tensor]:
        tensors = list()

        for decoderlayer, decodernorm in zip(self.decoderlayers, self.decodernorms):
            tensor = decodernorm(self.dropoutlayer(self.relu(decoderlayer(tensor))))
            tensors.append(tensor)

        reconstruction = self.outputlayer(tensor)

        # Decompose reconstruction to depths and tnf signal
        depths_out = reconstruction.narrow(1, 0, self.nsamples)
        tnf_out = reconstruction.narrow(1, self.nsamples, self.ntnf)

        # If multiple samples, apply softmax
        if self.nsamples > 1:
            depths_out = _softmax(depths_out, dim=1)

        return depths_out, tnf_out

    def forward(
        self, depths: Tensor, tnf: Tensor
    ) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        tensor = _torch.cat((depths, tnf), 1)
        mu, logsigma = self._encode(tensor)
        latent = self.reparameterize(mu, logsigma)
        depths_out, tnf_out = self._decode(latent)

        return depths_out, tnf_out, mu, logsigma

    def calc_loss(
        self,
        depths_in: Tensor,
        depths_out: Tensor,
        tnf_in: Tensor,
        tnf_out: Tensor,
        mu: Tensor,
        logsigma: Tensor,
        weights: Tensor,
    ) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        # If multiple samples, use cross entropy, else use SSE for abundance
        if self.nsamples > 1:
            # Add 1e-9 to depths_out to avoid numerical instability.
            ce = -((depths_out + 1e-9).log() * depths_in).sum(dim=1)
            ce_weight = (1 - self.alpha) / _log(self.nsamples)
        else:
            ce = (depths_out - depths_in).pow(2).sum(dim=1)
            ce_weight = 1 - self.alpha

        sse = (tnf_out - tnf_in).pow(2).sum(dim=1)
        kld = -0.5 * (1 + logsigma - mu.pow(2) - logsigma.exp()).sum(dim=1)
        sse_weight = self.alpha / self.ntnf
        kld_weight = 1 / (self.nlatent * self.beta)
        reconstruction_loss = ce * ce_weight + sse * sse_weight
        kld_loss = kld * kld_weight
        loss = (reconstruction_loss + kld_loss) * weights

        return loss.mean(), ce.mean(), sse.mean(), kld.mean()

    def trainepoch(
        self,
        data_loader: _DataLoader,
        epoch: int,
        optimizer,
        batchsteps: list[int],
        logfile,
    ) -> _DataLoader[tuple[Tensor, Tensor, Tensor]]:
        self.train()

        epoch_loss = 0.0
        epoch_kldloss = 0.0
        epoch_sseloss = 0.0
        epoch_celoss = 0.0

        if epoch in batchsteps:
            data_loader = set_batchsize(data_loader, data_loader.batch_size * 2)

        for depths_in, tnf_in, weights in data_loader:
            depths_in.requires_grad = True
            tnf_in.requires_grad = True

            if self.usecuda:
                depths_in = depths_in.cuda()
                tnf_in = tnf_in.cuda()
                weights = weights.cuda()

            optimizer.zero_grad()

            depths_out, tnf_out, mu, logsigma = self(depths_in, tnf_in)

            loss, ce, sse, kld = self.calc_loss(
                depths_in, depths_out, tnf_in, tnf_out, mu, logsigma, weights
            )

            loss.backward()
            optimizer.step()

            epoch_loss += loss.data.item()
            epoch_kldloss += kld.data.item()
            epoch_sseloss += sse.data.item()
            epoch_celoss += ce.data.item()

        if logfile is not None:
            print(
                "\tEpoch: {}\tLoss: {:.6f}\tCE: {:.7f}\tSSE: {:.6f}\tKLD: {:.4f}\tBatchsize: {}".format(
                    epoch + 1,
                    epoch_loss / len(data_loader),
                    epoch_celoss / len(data_loader),
                    epoch_sseloss / len(data_loader),
                    epoch_kldloss / len(data_loader),
                    data_loader.batch_size,
                ),
                file=logfile,
            )

            logfile.flush()

        self.eval()
        return data_loader

    def encode(self, data_loader) -> _np.ndarray:
        """Encode a data loader to a latent representation with VAE

        Input: data_loader: As generated by train_vae

        Output: A (n_contigs x n_latent) Numpy array of latent repr.
        """

        self.eval()

        new_data_loader = set_batchsize(
            data_loader, data_loader.batch_size, encode=True
        )

        depths_array, _, _ = data_loader.dataset.tensors
        length = len(depths_array)

        # We make a Numpy array instead of a Torch array because, if we create
        # a Torch array, then convert it to Numpy, Numpy will believe it doesn't
        # own the memory block, and array resizes will not be permitted.
        latent = _np.empty((length, self.nlatent), dtype=_np.float32)

        row = 0
        with _torch.no_grad():
            for depths, tnf, weights in new_data_loader:
                # Move input to GPU if requested
                if self.usecuda:
                    depths = depths.cuda()
                    tnf = tnf.cuda()

                # Evaluate
                _, _, mu, _ = self(depths, tnf)

                if self.usecuda:
                    mu = mu.cpu()

                latent[row : row + len(mu)] = mu
                row += len(mu)

        assert row == length
        return latent

    def save(self, filehandle):
        """Saves the VAE to a path or binary opened file. Load with VAE.load

        Input: Path or binary opened filehandle
        Output: None
        """
        state = {
            "nsamples": self.nsamples,
            "alpha": self.alpha,
            "beta": self.beta,
            "dropout": self.dropout,
            "nhiddens": self.nhiddens,
            "nlatent": self.nlatent,
            "state": self.state_dict(),
        }

        _torch.save(state, filehandle)

    @classmethod
    def load(
        cls, path: Union[IO[bytes], str], cuda: bool = False, evaluate: bool = True
    ):
        """Instantiates a VAE from a model file.

        Inputs:
            path: Path to model file as created by functions VAE.save or
                  VAE.trainmodel.
            cuda: If network should work on GPU [False]
            evaluate: Return network in evaluation mode [True]

        Output: VAE with weights and parameters matching the saved network.
        """

        # Forcably load to CPU even if model was saves as GPU model
        dictionary = _torch.load(path, map_location=lambda storage, loc: storage)

        nsamples = dictionary["nsamples"]
        alpha = dictionary["alpha"]
        beta = dictionary["beta"]
        dropout = dictionary["dropout"]
        nhiddens = dictionary["nhiddens"]
        nlatent = dictionary["nlatent"]
        state = dictionary["state"]

        vae = cls(nsamples, nhiddens, nlatent, alpha, beta, dropout, cuda)
        vae.load_state_dict(state)

        if cuda:
            vae.cuda()

        if evaluate:
            vae.eval()

        return vae

    def trainmodel(
        self,
        dataloader: _DataLoader[tuple[Tensor, Tensor, Tensor]],
        nepochs: int = 500,
        lrate: float = 1e-3,
        batchsteps: Optional[list[int]] = [25, 75, 150, 300],
        logfile: Optional[IO[str]] = None,
        modelfile: Union[None, str, Path, IO[bytes]] = None,
    ):
        """Train the autoencoder from depths array and tnf array.

        Inputs:
            dataloader: DataLoader made by make_dataloader
            nepochs: Train for this many epochs before encoding [500]
            lrate: Starting learning rate for the optimizer [0.001]
            batchsteps: None or double batchsize at these epochs [25, 75, 150, 300]
            logfile: Print status updates to this file if not None [None]
            modelfile: Save models to this file if not None [None]

        Output: None
        """

        if lrate < 0:
            raise ValueError(f"Learning rate must be positive, not {lrate}")

        if nepochs < 1:
            raise ValueError("Minimum 1 epoch, not {nepochs}")

        if batchsteps is None:
            batchsteps_set: set[int] = set()
        else:
            # First collect to list in order to allow all element types, then check that
            # they are integers
            batchsteps = list(batchsteps)
            if not all(isinstance(i, int) for i in batchsteps):
                raise ValueError("All elements of batchsteps must be integers")
            if max(batchsteps, default=0) >= nepochs:
                raise ValueError("Max batchsteps must not equal or exceed nepochs")
            last_batchsize = dataloader.batch_size * 2 ** len(batchsteps)
            if len(dataloader.dataset) < last_batchsize:  # type: ignore
                raise ValueError(
                    f"Last batch size of {last_batchsize} exceeds dataset length "
                    f"of {len(dataloader.dataset)}. "  # type: ignore
                    "This means you have too few contigs left after filtering to train. "
                    "It is not adviced to run Vamb with fewer than 10,000 sequences "
                    "after filtering. "
                    "Please check the Vamb log file to see where the sequences were "
                    "filtered away, and verify BAM files has sensible content."
                )
            batchsteps_set = set(batchsteps)

        # Get number of features
        # Following line is un-inferrable due to typing problems with DataLoader
        ncontigs, nsamples = dataloader.dataset.tensors[0].shape  # type: ignore
        optimizer = _Adam(self.parameters(), lr=lrate)

        if logfile is not None:
            print("\tNetwork properties:", file=logfile)
            print("\tCUDA:", self.usecuda, file=logfile)
            print("\tAlpha:", self.alpha, file=logfile)
            print("\tBeta:", self.beta, file=logfile)
            print("\tDropout:", self.dropout, file=logfile)
            print("\tN hidden:", ", ".join(map(str, self.nhiddens)), file=logfile)
            print("\tN latent:", self.nlatent, file=logfile)
            print("\n\tTraining properties:", file=logfile)
            print("\tN epochs:", nepochs, file=logfile)
            print("\tStarting batch size:", dataloader.batch_size, file=logfile)
            batchsteps_string = (
                ", ".join(map(str, sorted(batchsteps_set)))
                if batchsteps_set
                else "None"
            )
            print("\tBatchsteps:", batchsteps_string, file=logfile)
            print("\tLearning rate:", lrate, file=logfile)
            print("\tN sequences:", ncontigs, file=logfile)
            print("\tN samples:", nsamples, file=logfile, end="\n\n")

        # Train
        for epoch in range(nepochs):
            dataloader = self.trainepoch(
                dataloader, epoch, optimizer, sorted(batchsteps_set), logfile
            )

        # Save weights - Lord forgive me, for I have sinned when catching all exceptions
        if modelfile is not None:
            try:
                self.save(modelfile)
            except:
                pass

        return None



import unittest
import io
import numpy as np

import vamb
import testtools
from vamb.parsecontigs import CompositionMetaData


class TestParseBam(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        minlen = 3000
        mask = np.array(
            list(map(lambda x: x >= minlen, testtools.BAM_SEQ_LENS)), dtype=bool
        )
        cls.comp_metadata = CompositionMetaData(
            np.array(
                [i for (i, m) in zip(testtools.BAM_NAMES, mask) if m], dtype=object
            ),
            np.array([i for (i, m) in zip(testtools.BAM_SEQ_LENS, mask) if m]),
            mask,
            minlen,
        )

        cls.abundance = vamb.parsebam.Abundance.from_files(
            testtools.BAM_FILES, "/tmp/bam_tmpfile", cls.comp_metadata, True, 0.0, 2
        )

    def test_refhash(self):
        m = self.comp_metadata
        cp = CompositionMetaData(m.identifiers, m.lengths, m.mask, m.minlength)
        cp.refhash = b"a" * 32  # write bad refhash
        with self.assertRaises(ValueError):
            vamb.parsebam.Abundance.from_files(
                testtools.BAM_FILES, None, cp, True, 0.97, 4
            )

    def test_bad_metadata_mask(self):
        m = self.comp_metadata

        # If last element of mask is False, then the invariants of CompositionMetaData will
        # not hold after removing the last element of its mask, and that is NOT what we
        # are testing here.
        assert list(m.mask[-3:]) == [True, False, False]
        cp = CompositionMetaData(
            m.identifiers[:-1], m.lengths[:-1], m.mask[:-3], m.minlength
        )
        with self.assertRaises(ValueError):
            vamb.parsebam.Abundance.from_files(
                testtools.BAM_FILES, None, cp, True, 0.97, 4
            )

    def test_badfile(self):
        with self.assertRaises(FileNotFoundError):
            vamb.parsebam.Abundance.from_files(
                ["noexist"], None, self.comp_metadata, True, 0.97, 1
            )

    # Minid too high
    def test_minid_off(self):
        with self.assertRaises(ValueError):
            vamb.parsebam.Abundance.from_files(
                testtools.BAM_FILES, None, self.comp_metadata, True, 1.01, 4
            )

    def test_parse(self):
        nm = sum(self.comp_metadata.mask)
        self.assertEqual(nm, 12)

        self.assertEqual(self.abundance.matrix.shape, (nm, 3))
        self.assertEqual(self.abundance.nseqs, nm)
        self.assertEqual(self.abundance.matrix.dtype, np.float32)

    def test_minid(self):
        abundance = vamb.parsebam.Abundance.from_files(
            testtools.BAM_FILES, None, self.comp_metadata, True, 0.95, 3
        )
        self.assertTrue(np.any(abundance.matrix < self.abundance.matrix))

    def test_save_load(self):
        buf = io.BytesIO()
        self.abundance.save(buf)
        buf.seek(0)

        # Bad refhash
        with self.assertRaises(ValueError):
            abundance2 = vamb.parsebam.Abundance.load(buf, b"a" * 32)

        buf.seek(0)
        abundance2 = vamb.parsebam.Abundance.load(buf, self.abundance.refhash)
        self.assertTrue(np.all(abundance2.matrix == self.abundance.matrix))
        self.assertTrue(np.all(abundance2.samplenames == self.abundance.samplenames))
        self.assertEqual(abundance2.refhash, self.abundance.refhash)
        self.assertEqual(abundance2.minid, self.abundance.minid)



"""Vamb - Variational Autoencoders for Metagenomic Binning
Documentation: https://github.com/RasmussenLab/vamb/

Vamb contains the following modules:
vamb.vambtools
vamb.parsecontigs
vamb.parsebam
vamb.encode
vamb.cluster
vamb.benchmark

General workflow:
1) Filter contigs by size using vamb.vambtools.filtercontigs
2) Map reads to contigs to obtain BAM files
3) Calculate a Composition of contigs using vamb.parsecontigs
4) Create Abundance object from BAM files using vamb.parsebam
5) Train autoencoder using vamb.encode
6) Cluster latent representation using vamb.cluster
7) Split bins using vamb.vambtools
"""

__version__ = (4, 1, 2)

from . import vambtools
from . import parsebam
from . import parsecontigs
from . import cluster
from . import benchmark
from . import encode
from . import aamb_encode

__all__ = [
    "vambtools",
    "parsebam",
    "parsecontigs",
    "cluster",
    "benchmark",
    "encode",
    "aamb_encode",
]



# The approach to benchmarking is this:

# A "subject" is a source sequence, e.g. a sequence directly from an underlying genome.
# A subject has only a name and length. Subject names are unique.

# A set of subjects is a Genome. The subjects are assumed to not be overlapping. Hence,
# the "breadth" of a genome is the sum of subject lengths in that genome. Genome names
# are assumed to be unique.

# A Contig is composed of a name, a source subject, and a start and end position within
# that subject, where the contig maps to. Contig names are assumed to be unique.
# Contigs always have a source. If the source is unknown, a dummy source can be used,
# i.e. a source not used by any other contig.

# A named set of Contigs is a Bin. Bin names are unique. The "breadth" of a Bin is the total
# number of basepairs (positions) covered by any contig in the bin. Contigs can overlap the
# same subject, but a position covered multiple times only count once towards breadth.

# A Bin/Genome pair has an "intersection". This is the number of basepairs (positions) in
# the Bin which come from a subject in the given Genome. In other words, it is the breadth
# of the subset of contigs in the Bin that originates from the Genome.

# Bin/Genome pairs also have other statistics:
# * True positives (TP) is the intersection
# * False positives (FP) is Bin breadth minus TP
# * False negatives (FN) is Genome breadth minus TP
# * Recall, precision and F-score is calculated from TP, FP and FN as usual.

# A binning result is described by two objects, the Reference and the Binning.

# A Reference object represents the information in the test dataset, and is independent of
# any concrete binning tool or binning result. Hence, a test dataset has one Reference object.
# It is composed of:
# * The total set of Contigs input to the binning
# * The total set of Genomes that the set of Contigs represent. This implies that all Contigs
#   in the Reference must derive from a subject which is also in a Genome in the Reference, i.e.
#   there must be no contigs where the source is not from a Genome (via a subject) in the Reference.
# * A list of "taxmaps". A taxmap describe taxonomic information about the Genomes of the
#   Reference. A taxmap is a dictionary that maps the name of a clade to its parent clade.
#   For example, the pair "Ecoli":"Escherichia" could be present.
#   The first taxmap has all Genome names as keys.
#   The Nth taxmap has as keys the values of the N-1th taxmap (except None).
#   A clade for which no higher level clade is known or defined maps to None.

# The Binning object contains the matching Reference and a collection of Bins.
# When instantiated, it can do a number of filters or transformations of the bins, e.g.
# binsplitting or filtering of bins by number of contigs.
# The only responsibility of the Binning object is to store the input bins, do this filtering
# and transformation, and do the benchmarking itself. In practice, this means computing a
# series of "counters", one for each Reference taxmap:

# The first counter is a map from minimum (recall, precision) to the number of genomes
# for which at least one bin in the Binning has a recall, precision level at or above the
# minimum recall/precision level.

# All the other counters are computed iteratively from the previous one based on the corresponding
# Reference taxmap. For a given bin B and a clade C where C  is not a Genome (i.e. not the lowest
# level of the taxmaps), the recall of the B/C pair is defined as the max of the recalls B/Cchild for
# all Cchild taxonomic children of C according to the taxmap.
# The precision is calculated as the SUM of all B/Cchild.


__doc__ = """Benchmark script

This benchmarks bins using number of covered sites.

First, a Reference is needed. This can be loaded from a JSON file, see
the Reference class source code for details.

You also need a file with the binning. This is simply the clusters.tsv file
as produced by Vamb, i.e. first column is the name of the bin of a contig,
second column the name of the contig.

Recall of a genome/bin pair is defined as the number of genome basepairs
covered by contigs in the bin divided by total number of basepairs of that
genome.
Precision is the number of bases in that genome covered by contigs in the
bin divided by the number of bases covered in any genome from that bin.

Usage:
>>> ref = Reference.from_file(open_reference_file_hande)
>>> bins = Binning.from_file(open_clusters_file_handle, ref)
>>> bins.print_matrix(rank=1)
"""

from collections import defaultdict
from itertools import product
import sys
import json
from math import isfinite
from vamb import vambtools
from collections.abc import Iterable, Sequence
from typing import Optional, TypeVar, IO, Any

C = TypeVar("C", bound="Contig")
G = TypeVar("G", bound="Genome")
Bn = TypeVar("Bn", bound="Bin")
R = TypeVar("R", bound="Reference")
Bs = TypeVar("Bs", bound="Binning")


class Contig:
    """An object representing a contig mapping to a subject at position start:end.
    Mapping positions use the half-open interval, like Python ranges and slices.

    Instantiate either with name, subject and mapping start/end:
        Contig('contig_41', 'subject_1', 600, 11, 510)
    Or with only name and length
        Contig.subjectless('contig_41', 499)
    A subjectless Contig uses itself as a subject (implying it only maps to itself).
    """

    __slots__ = ["name", "subject", "start", "end"]

    def __init__(self, name: str, subject: str, start: int, end: int):
        if start < 0:
            raise ValueError(f'Contig "{name}" has negative start index {start}')
        elif end <= start:
            raise ValueError(
                "Contig end must be higher than start, but "
                f'contig "{name}" spans {start}-{end}.'
            )

        self.name = name
        self.subject = subject
        self.start = start
        self.end = end

    @classmethod
    def subjectless(cls: type[C], name: str, length: int) -> C:
        "Instantiate with only name and length"
        return cls(name, name, 0, length)

    def __repr__(self) -> str:
        return f'Contig("{self.name}", "{self.subject}", {self.start}, {self.end})'

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Contig) and self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name) ^ 3458437981

    def __len__(self) -> int:
        return self.end - self.start


class Genome:
    """An object representing a set of subjects (or sources), i.e. genomes or source contigs
    that the binning contigs are drawn from.
    >>> g = Genome("Ecoli")
    >>> g.add("chrom", 5_300_000)
    """

    __slots__ = ["name", "sources", "breadth"]

    def __init__(self, name: str):
        self.name = name
        self.sources: dict[str, int] = dict()
        self.breadth = 0

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Genome) and self.name == other.name

    def __hash__(self) -> int:
        return hash(self.name) ^ 4932511201

    def add(self, source: str, len: int) -> None:
        if len <= 0:
            raise ValueError(
                "Source sequence must have nonzero positive length, "
                f"but {source} has length {len}"
            )
        if source in self.sources:
            raise ValueError(f"Genome {self.name} already has source {source}.")

        self.sources[source] = len
        self.breadth += len

    def __repr__(self):
        return f'Genome("{self.name}")'


class Bin:
    """An object representing a set of Contigs.
    Should be instantiated with Bin.from_contigs. See that method for how to safely
    instantiate the object.
    """

    __slots__ = ["name", "contigs", "intersections", "breadth"]

    def __init__(self, name: str) -> None:
        self.name = name
        self.contigs: set[Contig] = set()
        self.intersections: dict[Genome, int] = dict()
        self.breadth: int = 0

    @property
    def ncontigs(self) -> int:
        return len(self.contigs)

    @classmethod
    def from_contigs(
        cls: type[Bn],
        name: str,
        contigs: Iterable[Contig],
        genomeof: dict[Contig, Genome],
    ) -> Bn:
        instance = cls(name)
        instance.contigs.update(contigs)
        instance._finalize(genomeof)  # remember to do this
        return instance

    def __repr__(self) -> str:
        return f'Bin("{self.name}")'

    @staticmethod
    def _intersection(contigs: list[Contig]) -> int:
        # Note: All contigs MUST come from same source, else the code is invalid
        contigs.sort(key=lambda x: x.start)
        result = 0
        rightmost_end = -1

        for contig in contigs:
            result += max(contig.end, rightmost_end) - max(contig.start, rightmost_end)
            rightmost_end = max(contig.end, rightmost_end)

        return result

    def _finalize(self, genomeof: dict[Contig, Genome]) -> None:
        self.intersections.clear()
        by_source: defaultdict[tuple[Genome, str], list[Contig]] = defaultdict(list)
        for contig in self.contigs:
            by_source[(genomeof[contig], contig.subject)].append(contig)
        for (genome, _), contigs in by_source.items():
            self.intersections[genome] = self.intersections.get(
                genome, 0
            ) + self._intersection(contigs)
        self.breadth = sum(self.intersections.values())

    def confusion_matrix(self, genome: Genome) -> tuple[int, int, int]:
        "Given a genome and a binname, returns TP, FP, FN"
        # This is None if it's not updated, we want it to throw type errors then
        d: dict[Genome, int] = self.intersections
        breadth: int = self.breadth
        tp = d.get(genome, 0)
        fp = breadth - tp
        fn = genome.breadth - tp
        return (tp, fp, fn)

    def recall_precision(self, genome: Genome) -> tuple[float, float]:
        (tp, fp, fn) = self.confusion_matrix(genome)
        recall = tp / (tp + fn)
        precision = tp / (tp + fp)
        return (recall, precision)

    def fscore(self, b: float, genome: Genome) -> float:
        recall, precision = self.recall_precision(genome)

        # Some people say the Fscore is undefined in this case.
        # We define it to be 0.0
        if (recall + precision) == 0.0:
            return 0.0

        return (1 + b * b) * (recall * precision) / ((b * b * precision) + recall)

    def f1(self, genome: Genome) -> float:
        return self.fscore(1.0, genome)


class Reference:
    """An object that represent a set of Genome and Contigs, where the Contigs are sampled
    from the genomes. Also contain the phylogenetic tree for the contained genomes.
    Either instantiate directly and use self.add_contig and self.add_taxonomy, else
    use Reference.from_file
    """

    __slots__ = ["genomes", "genomeof", "contig_by_name", "taxmaps"]

    def __init__(self):
        self.genomes: set[Genome] = set()  # genome_name : genome dict
        self.genomeof: dict[Contig, Genome] = dict()  # contig : genome dict
        self.contig_by_name: dict[str, Contig] = dict()
        # This is a list of dicts: The first one maps genomename to name of next taxonomic level
        # The second maps name of second level to name of third level etc. None means it's the top level
        self.taxmaps: list[dict[str, Optional[str]]] = [dict()]

    def _add_genome(self, genome: Genome) -> None:
        if genome in self.genomes:
            raise ValueError(f'Genome "{genome.name}" already in reference.')
        self.genomes.add(genome)
        self.taxmaps[0][genome.name] = None

    def _add_contig(self, contig: Contig, genome: Genome) -> None:
        if contig in self.genomeof:
            raise ValueError(f'Reference already has Contig "{contig.name}"')
        if contig.subject not in genome.sources:
            raise ValueError(
                f'Attempted to add contig "{contig.name}" with source "{contig.subject}" '
                f'to genome "{genome.name}", but genome has no such source'
            )
        if genome not in self.genomes:
            raise ValueError(f'Genome "{genome.name}" is not in reference.')
        if genome.sources[contig.subject] < contig.end:
            raise IndexError(
                f'Attempted to add contig "{contig.name}" with mapping end {contig.end} '
                f'to subject "{contig.subject}", '
                f"but the subject only has length {genome.sources[contig.subject]}"
            )

        self.genomeof[contig] = genome
        self.contig_by_name[contig.name] = contig

    def _add_taxonomy(self, level: int, child: str, parent: str) -> None:
        existing = self.taxmaps[level][child]
        if existing is None:
            self.taxmaps[level][child] = parent
            # Add parent of parent, if not already present
            if level + 1 == self.nranks:
                self.taxmaps.append(dict())

            grandparent = self.taxmaps[level + 1].get(parent, 0)
            if grandparent == 0:
                self.taxmaps[level + 1][parent] = None

        elif existing != parent:
            raise ValueError(
                f'Clade "{child}" maps to both parent clade "{parent}" and "{existing}"'
            )

    @property
    def ngenomes(self) -> int:
        return len(self.genomes)

    @property
    def ncontigs(self) -> int:
        return len(self.genomeof)

    @property
    def nranks(self) -> int:
        return len(self.taxmaps)

    def __repr__(self) -> str:
        return f"<Reference with {self.ngenomes} genomes, {self.ncontigs} contigs and {self.nranks} ranks>"

    def parse_bins(
        self, io: Iterable[str], binsplit_sep: Optional[str] = None
    ) -> list[Bin]:
        clusters = vambtools.read_clusters(io).items()
        if binsplit_sep is not None:
            clusters = vambtools.binsplit(clusters, binsplit_sep)
        return self.load_bins(clusters)

    def load_bins(self, bins: Iterable[tuple[str, Iterable[str]]]) -> list[Bin]:
        """Convert a set of bin names to a list of Bins"""
        result: list[Bin] = list()
        for binname, contignames in bins:
            contigs = (self.contig_by_name[name] for name in contignames)
            result.append(Bin.from_contigs(binname, contigs, self.genomeof))

        return result

    @classmethod
    def from_file(cls: type[R], io: IO[str]) -> R:
        json_dict = json.load(io)
        return cls.from_dict(json_dict)

    @classmethod
    def from_dict(cls: type[R], json_dict: dict[str, Any]) -> R:
        instance = cls()
        for genomename, sourcesdict in json_dict["genomes"].items():
            genome = Genome(genomename)
            instance._add_genome(genome)
            for sourcename, (sourcelen, contigdict) in sourcesdict.items():
                genome.add(sourcename, sourcelen)
                for contigname, (start, end) in contigdict.items():
                    # JSON format is 1-indexed and includes endpoints, whereas
                    # Contig struct is not, so compensate.
                    contig = Contig(contigname, sourcename, start - 1, end)
                    instance._add_contig(contig, genome)

        for _ in range(len(json_dict["taxmaps"]) - 1):
            instance.taxmaps.append(dict())
        for level, taxmap in enumerate(json_dict["taxmaps"]):
            for child, parent in taxmap.items():
                instance._add_taxonomy(level, child, parent)

        return instance

    def save(self, io: IO[str]) -> None:
        json_dict: dict[str, Any] = {"genomes": dict(), "taxmaps": []}

        # "genomes": {genomename: [subject_len, {contigname: [start, stop]}]}
        genome_dict = json_dict["genomes"]
        for genome in self.genomes:
            source_dict: dict[str, list[Any]] = dict()
            genome_dict[genome.name] = source_dict
            for sourcename, length in genome.sources.items():
                source_dict[sourcename] = [length, dict()]

        for contig, genome in self.genomeof.items():
            # JSON format is 1-indexes and includes endpoints, whereas
            # Contig struct is not, so compensate.
            genome_dict[genome.name][contig.subject][1][contig.name] = [
                contig.start + 1,
                contig.end,
            ]

        for taxmap in self.taxmaps:
            d: dict[str, str] = dict()
            for child, parent in taxmap.items():
                if parent is not None:
                    d[child] = parent
            if len(d) > 0:
                json_dict["taxmaps"].append(d)

        json.dump(json_dict, io)


class Binning:
    """The result of a set of Bins applied to a Reference.
    See Binning.from_file for more usage.
    >>> with open("clusters.tsv") as file:
    ...     binning = Binning.from_file(file, reference)

    Properties:
    * reference: Reference
    * bins: list[Bin]

    Properties after self.benchmark()
    * counters: list[dict[tuple[float, float], int]]: Genomes at recall/prec thresholds
    * recalls: Recall thresholds used to compute counters
    * precisions: Precision thresholds used to compute counters

    Extra arguments to Binning.from_file:
    * disjoint: If True, do not allow same contig in multiple bins
    * binsplit_separator: If str and not None, split bins by separator
    * minsize: Filter away all bins with breadth less than this
    * mincontigs: Filter away bins with fewer contigs than this
    """

    __slots__ = ["reference", "bins", "counters", "recalls", "precisions"]
    _DEFAULTRECALLS = (0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.99)
    _DEFAULTPRECISIONS = (0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.99)

    def __init__(
        self,
        bins: Sequence[Bin],
        reference: Reference,
        recalls: Sequence[float] = _DEFAULTRECALLS,
        precisions: Sequence[float] = _DEFAULTPRECISIONS,
        disjoint: bool = True,
    ) -> None:
        self.recalls: tuple[float] = self._validate_rec_prec(recalls)
        self.precisions: tuple[float] = self._validate_rec_prec(precisions)
        self.reference = reference
        self.bins: list[Bin] = list(bins)
        # This is initialized by self._benchmark
        self.counters: list[dict[tuple[float, float], int]] = list()

        if disjoint:
            seen_contigs: set[Contig] = set()
            for bin in self.bins:
                for contig in bin.contigs:
                    if contig in seen_contigs:
                        raise ValueError(
                            f'Contig "{contig.name}" seen twice in disjoint binning.'
                        )
                    seen_contigs.add(contig)
        self._benchmark()

    @classmethod
    def from_file(
        cls: type[Bs],
        filehandle: Iterable[str],
        reference: Reference,
        recalls: Sequence[float] = _DEFAULTRECALLS,
        precisions: Sequence[float] = _DEFAULTPRECISIONS,
        disjoint: bool = True,
        binsplit_separator: Optional[str] = None,
        minsize: int = 1,
        mincontigs: int = 1,
    ) -> Bs:
        bins = reference.parse_bins(filehandle, binsplit_separator)
        bins = cls.filter_bins(bins, minsize, mincontigs)
        instance = cls(bins, reference, recalls, precisions, disjoint)
        return instance

    @classmethod
    def gold_standard(
        cls: type[Bs],
        reference: Reference,
        recalls: Sequence[float] = _DEFAULTRECALLS,
        precisions: Sequence[float] = _DEFAULTPRECISIONS,
    ) -> Bs:
        "Return a Binning from a given Reference where each Genome is precisely one Bin"
        contigsof: defaultdict[Genome, list[Contig]] = defaultdict(list)
        for contig in reference.contig_by_name.values():
            contigsof[reference.genomeof[contig]].append(contig)
        bins = [
            Bin.from_contigs(genome.name, contigs, reference.genomeof)
            for (genome, contigs) in contigsof.items()
        ]
        instance = cls(bins, reference, recalls, precisions, disjoint=False)
        return instance

    def print_matrix(self, rank: int, file: IO[str] = sys.stdout) -> None:
        """Prints the recall/precision number of bins to STDOUT."""
        recalls: tuple[float] = self.recalls
        precisions: tuple[float] = self.precisions
        assert self.counters is not None

        if rank >= self.reference.nranks:
            raise IndexError("Taxonomic rank out of range")

        print("\tRecall", file=file)
        print("Prec.", "\t".join([str(r) for r in recalls]), sep="\t", file=file)

        for min_precision in precisions:
            row = [
                self.counters[rank][(min_recall, min_precision)]
                for min_recall in recalls
            ]
            print(min_precision, "\t".join([str(i) for i in row]), sep="\t", file=file)

    def __repr__(self) -> str:
        return (
            f"<Binning with {self.nbins} bins and reference {hex(id(self.reference))}>"
        )

    @property
    def nbins(self) -> int:
        return len(self.bins)

    @staticmethod
    def filter_bins(bins: Iterable[Bin], minsize: int, mincontigs: int) -> list[Bin]:
        def is_ok(bin: Bin) -> bool:
            breadth: int = bin.breadth
            return breadth >= minsize and bin.ncontigs >= mincontigs

        return list(filter(is_ok, bins))

    def _benchmark(self) -> None:
        counters: list[dict[tuple[float, float], int]] = list()
        # key here is name of genome (and later, other taxonomic ranks)
        rp_by_name: dict[str, dict[Bin, tuple[float, float]]] = {
            g.name: dict() for g in self.reference.genomes
        }
        for bin in self.bins:
            assert bin.intersections is not None
            for genome in bin.intersections:
                rp_by_name[genome.name][bin] = bin.recall_precision(genome)
        bitvectors = self._get_seen_bitvectors(rp_by_name)
        counters.append(self._counter_from_bitvectors(bitvectors))

        for rank in range(self.reference.nranks - 1):
            rp_by_name = self._uprank_rp_by_name(rank, rp_by_name)
            bitvectors = self._get_seen_bitvectors(rp_by_name)
            counters.append(self._counter_from_bitvectors(bitvectors))

        self.counters = counters

    @staticmethod
    def _validate_rec_prec(x: Iterable[float]) -> tuple[float]:
        s: set[float] = set()
        for i in x:
            if i in s:
                raise ValueError(f"Recall/precision value {i} present multiple times.")
            if not isfinite(i) or i <= 0.0 or i > 1:
                raise ValueError(
                    f"Recall/precision value {i} is not a finite value in (0;1]"
                )
            s.add(i)
        if len(s) == 0:
            raise ValueError("Must provide at least 1 recall/precision value")
        return tuple(sorted(s))

    def _get_seen_bitvectors(
        self, rp_by_name: dict[str, dict[Bin, tuple[float, float]]]
    ) -> dict[str, int]:
        recalls: tuple[float] = self.recalls
        precisions: tuple[float] = self.precisions
        bitvectors: dict[str, int] = dict()

        for cladename, d in rp_by_name.items():
            bitvector = 0
            for recall, precision in d.values():
                for i, (min_recall, min_precision) in enumerate(
                    product(recalls, precisions)
                ):
                    if recall >= min_recall and precision >= min_precision:
                        bitvector |= 1 << i
                    # Shortcut: Once we pass min recall, rest of values can be skipped
                    elif recall < min_recall:
                        break

            bitvectors[cladename] = bitvector

        return bitvectors

    def _counter_from_bitvectors(
        self, bitvectors: dict[str, int]
    ) -> dict[tuple[float, float], int]:
        recalls: tuple[float] = self.recalls
        precisions: tuple[float] = self.precisions
        result: dict[tuple[float, float], int] = {
            (r, p): 0 for (r, p) in product(recalls, precisions)
        }
        for _, bitvector in bitvectors.items():
            for i, rp in enumerate(product(recalls, precisions)):
                result[rp] += (bitvector >> i) & 1

        return result

    def _uprank_rp_by_name(
        self, fromrank: int, rp_by_name: dict[str, dict[Bin, tuple[float, float]]]
    ) -> dict[str, dict[Bin, tuple[float, float]]]:
        result: dict[str, dict[Bin, tuple[float, float]]] = dict()
        for child, parent in self.reference.taxmaps[fromrank].items():
            # If no parent clade, we just name the "parent" same as child
            if parent is None:
                parent = child

            if parent not in result:
                result[parent] = dict()

            parent_dict = result[parent]
            for bin, (old_recall, old_prec) in rp_by_name[child].items():
                (new_recall, new_prec) = parent_dict.get(bin, (0.0, 0.0))
                parent_dict[bin] = (max(old_recall, new_recall), old_prec + new_prec)

        return result



import unittest
import numpy as np
import torch
import tempfile
import io

import vamb


class TestDataLoader(unittest.TestCase):
    tnfs = np.random.random((111, 103)).astype(np.float32)
    rpkm = np.random.random((111, 14)).astype(np.float32)
    lens = np.random.randint(2000, 5000, size=111)

    def nearly_same(self, A, B):
        self.assertTrue(np.all(np.abs(A - B) < 1e-5))

    def not_nearly_same(self, A, B):
        self.assertTrue(np.any(np.abs(A - B) > 1e-4))

    def test_bad_args(self):
        # Bad rpkm
        with self.assertRaises(ValueError):
            vamb.encode.make_dataloader(
                [[1, 2, 3]], self.tnfs, np.ndarray([2000]), batchsize=32
            )

        # bad tnfs
        with self.assertRaises(ValueError):
            vamb.encode.make_dataloader(
                self.rpkm, [[1, 2, 3]], np.ndarray([2000]), batchsize=32
            )

        # Bad batchsize
        with self.assertRaises(ValueError):
            vamb.encode.make_dataloader(self.rpkm, self.tnfs, self.lens, batchsize=0)

        # Differing lengths
        with self.assertRaises(ValueError):
            vamb.encode.make_dataloader(
                np.random.random((len(self.rpkm) - 1)).astype(np.float32),
                self.tnfs,
                self.lens,
                batchsize=32,
            )

        # Bad dtype
        with self.assertRaises(ValueError):
            vamb.encode.make_dataloader(
                self.rpkm.astype(np.float64), self.tnfs, self.lens, batchsize=32
            )

        # Batchsize is too large
        with self.assertRaises(ValueError):
            vamb.encode.make_dataloader(self.rpkm, self.tnfs, self.lens, batchsize=256)

    def test_destroy(self):
        copy_rpkm = self.rpkm.copy()
        copy_tnfs = self.tnfs.copy()

        (dl, mask) = vamb.encode.make_dataloader(
            self.rpkm, self.tnfs, self.lens, batchsize=32
        )
        self.nearly_same(self.rpkm, copy_rpkm)
        self.nearly_same(self.tnfs, copy_tnfs)

        (dl, mask) = vamb.encode.make_dataloader(
            copy_rpkm, copy_tnfs, self.lens, batchsize=32, destroy=True
        )
        self.not_nearly_same(self.rpkm, copy_rpkm)
        self.not_nearly_same(self.tnfs, copy_tnfs)

    def test_normalized(self):
        copy_rpkm = self.rpkm.copy()
        copy_tnfs = self.tnfs.copy()

        (dl, mask) = vamb.encode.make_dataloader(
            copy_rpkm, copy_tnfs, self.lens, batchsize=32, destroy=True
        )

        # TNFS: Mean of zero, std of one
        self.nearly_same(np.mean(copy_tnfs, axis=0), np.zeros(copy_tnfs.shape[1]))
        self.nearly_same(np.std(copy_tnfs, axis=0), np.ones(copy_tnfs.shape[1]))

        # RPKM: Sum to 1, all zero or above
        # print(copy_rpkm)
        self.nearly_same(np.sum(copy_rpkm, axis=1), np.ones(copy_rpkm.shape[0]))
        self.assertTrue(np.all(copy_rpkm >= 0.0))

    def test_mask(self):
        copy_rpkm = self.rpkm.copy()
        copy_tnfs = self.tnfs.copy()
        mask = np.ones(len(copy_rpkm)).astype(bool)

        for bad_tnf in [0, 4, 9]:
            copy_tnfs[bad_tnf, :] = 0
            mask[bad_tnf] = False

        for bad_rpkm in [1, 4, 11, 19]:
            copy_rpkm[bad_rpkm, :] = 0
            mask[bad_rpkm] = False

        (dl, mask2) = vamb.encode.make_dataloader(
            copy_rpkm, copy_tnfs, self.lens, batchsize=32
        )

        self.assertTrue(np.all(mask == mask2))

    def test_single_sample(self):
        single_rpkm = self.rpkm[:, [0]]
        (dl, mask) = vamb.encode.make_dataloader(
            single_rpkm, self.tnfs.copy(), self.lens, batchsize=32, destroy=True
        )
        self.assertLess(np.abs(np.mean(single_rpkm)), 1e-6)
        self.assertLess(abs(np.std(single_rpkm) - 1), 1e-6)

    def test_iter(self):
        bs = 32
        (dl, mask) = vamb.encode.make_dataloader(
            self.rpkm, self.tnfs, self.lens, batchsize=bs
        )

        # Check right element type
        for M in next(iter(dl)):
            self.assertEqual(M.dtype, torch.float32)
            self.assertEqual(M.shape[0], bs)

        # Check it iterates the right order (rpkm, tnfs)
        rpkm, tnfs, weights = next(iter(dl))
        self.nearly_same(np.sum(rpkm.numpy(), axis=1), np.ones(bs))

    def test_randomized(self):
        (dl, mask) = vamb.encode.make_dataloader(
            self.rpkm, self.tnfs, self.lens, batchsize=64
        )
        rpkm, tnfs, weights = next(iter(dl))

        # Test that first batch is not just the first 64 elements.
        # Could happen, but vanishingly unlikely.
        self.assertTrue(np.any(np.abs(tnfs.numpy() - self.tnfs[:64]) > 1e-3))


class TestVAE(unittest.TestCase):
    tnfs = np.random.random((111, 103)).astype(np.float32)
    rpkm = np.random.random((111, 14)).astype(np.float32)
    lens = np.random.randint(2000, 5000, size=111)

    def test_bad_args(self):
        with self.assertRaises(ValueError):
            vamb.encode.VAE(-1)

        with self.assertRaises(ValueError):
            vamb.encode.VAE(5, nlatent=0)

        with self.assertRaises(ValueError):
            vamb.encode.VAE(5, nhiddens=[128, 0])

        with self.assertRaises(ValueError):
            vamb.encode.VAE(5, alpha=0.0)

        with self.assertRaises(ValueError):
            vamb.encode.VAE(5, alpha=1.0)

        with self.assertRaises(ValueError):
            vamb.encode.VAE(5, beta=0.0)

        with self.assertRaises(ValueError):
            vamb.encode.VAE(5, dropout=1.0)

        with self.assertRaises(ValueError):
            vamb.encode.VAE(5, dropout=-0.001)

    def test_loss_falls(self):
        vae = vamb.encode.VAE(self.rpkm.shape[1])
        rpkm_copy = self.rpkm.copy()
        tnfs_copy = self.tnfs.copy()
        dl, mask = vamb.encode.make_dataloader(
            rpkm_copy, tnfs_copy, self.lens, batchsize=16, destroy=True
        )
        di = torch.Tensor(rpkm_copy)
        ti = torch.Tensor(tnfs_copy)
        we = dl.dataset.tensors[2]
        do, to, mu, lsigma = vae(di, ti)
        start_loss = vae.calc_loss(di, do, ti, to, mu, lsigma, we)[0].data.item()
        iobuffer = io.StringIO()

        with tempfile.TemporaryFile() as file:
            # Loss drops with training
            vae.trainmodel(
                dl, nepochs=3, batchsteps=[1, 2], logfile=iobuffer, modelfile=file
            )
            do, to, mu, lsigma = vae(di, ti)
            end_loss = vae.calc_loss(di, do, ti, to, mu, lsigma, we)[0].data.item()
            self.assertLess(end_loss, start_loss)

            # Also test save/load
            before_encoding = vae.encode(dl)
            file.flush()
            file.seek(0)
            vae_2 = vamb.encode.VAE.load(file)

        after_encoding = vae_2.encode(dl)
        self.assertTrue(np.all(np.abs(before_encoding - after_encoding) < 1e-6))

    def test_encoding(self):
        nlatent = 15
        vae = vamb.encode.VAE(self.rpkm.shape[1], nlatent=nlatent)
        dl, mask = vamb.encode.make_dataloader(
            self.rpkm, self.tnfs, self.lens, batchsize=32
        )
        encoding = vae.encode(dl)
        self.assertEqual(encoding.dtype, np.float32)
        self.assertEqual(encoding.shape, (len(self.rpkm), nlatent))
